"""
Document Store on SQLite.
"""

from __future__ import annotations

__all__ = ["SQLite"]

import copy
import json
import re
import sqlite3
from typing import Any

from x8.core import Context, DataAccessor, NCall, Operation, Response
from x8.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from x8.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    QueryFunctionName,
    Select,
    Update,
    UpdateOp,
    Value,
)
from x8.storage._common import (
    ArrayIndex,
    AscIndex,
    CollectionResult,
    CollectionStatus,
    CompositeIndex,
    DescIndex,
    FieldIndex,
    HashIndex,
    Index,
)
from x8.storage._common import IndexHelper as BaseIndexHelper
from x8.storage._common import (
    IndexResult,
    IndexStatus,
    ItemProcessor,
    ParameterParser,
    RangeIndex,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)

from .._feature import DocumentStoreFeature
from .._helper import (
    build_item_from_parts,
    build_item_from_value,
    build_query_result,
    get_collection_config,
)
from .._models import DocumentCollectionConfig, DocumentFieldType


class SQLite(StoreProvider):
    database: str
    table: str | None
    id_column: str | dict
    value_column: str | dict
    pk_column: str | dict | None
    id_map_field: str | dict | None
    pk_map_field: str | dict | None
    etag_embed_field: str | dict | None
    suppress_fields: list[str] | None
    nparams: dict[str, Any]

    _client: Any
    _aclient: Any
    _collection_cache: dict[str, SQLiteCollection]
    _acollection_cache: dict[str, SQLiteCollection]

    def __init__(
        self,
        database: str = ":memory:",
        table: str | None = None,
        id_column: str | dict = "id",
        value_column: str | dict = "value",
        pk_column: str | dict | None = "pk",
        id_map_field: str | dict | None = "id",
        pk_map_field: str | dict | None = "pk",
        etag_embed_field: str | dict | None = "_etag",
        suppress_fields: list[str] | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            database:
                SQLite database, defaults to ":memory:".
            table:
                SQLite table name mapped to document store collection.
            id_column:
                Column name to store the document id, defaults to "id".
            value_column:
                Column name to store the document store value.
                This column should have the JSON or JSONB type,
                defaults to "value".
            pk_column:
                Column name to store the document store pk,
                defaults to "pk".
            id_map_field:
                Field in the document to map into id.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field, defaults to "id".
            pk_map_field:
                Field in the document to map into pk.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field, defaults to "pk".
            etag_embed_field:
                Field to store the generated ETAG value.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field, defaults to "_etag".
            suppress_fields:
                List of fields to supress when results are returned.
            nparams:
                Native parameters to sqlite client.
        """
        self.database = database
        self.table = table
        self.id_column = id_column
        self.value_column = value_column
        self.pk_column = pk_column
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field
        self.etag_embed_field = etag_embed_field
        self.suppress_fields = suppress_fields
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._collection_cache = dict()
        self._acollection_cache = dict()

    def __supports__(self, feature: str) -> bool:
        return feature not in [DocumentStoreFeature.TYPE_BINARY]

    def __setup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        self._client = sqlite3.connect(
            self.database,
            check_same_thread=False,
            **self.nparams,
        )

    def _get_table_name(self, op_parser: StoreOperationParser) -> str:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        table = collection_name or self.table or self.__component__.collection
        if table is None:
            raise BadRequestError("Collection name must be specified.")
        return table

    def _get_collections(
        self, op_parser: StoreOperationParser
    ) -> list[SQLiteCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[SQLiteCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(self._get_collections(single_op_parser))
            return collections
        table = self._get_table_name(op_parser)
        if table is None:
            raise BadRequestError("Collection name must be specified")
        if table in self._collection_cache:
            return [self._collection_cache[table]]
        id_column = ParameterParser.get_collection_parameter(
            self.id_column, table
        )
        value_column = ParameterParser.get_collection_parameter(
            self.value_column, table
        )
        pk_column = ParameterParser.get_collection_parameter(
            self.pk_column, table
        )
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field or self.__component__.id_map_field, table
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field or self.__component__.pk_map_field, table
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, table
        )
        col = SQLiteCollection(
            table,
            id_column,
            value_column,
            pk_column,
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._collection_cache[table] = col
        return [col]

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[StoreOperation.PUT, StoreOperation.DELETE],
                single_collection=True,
            )

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collections = self._get_collections(op_parser)
        ncall, state = self._get_ncall(
            op_parser, collections, ClientHelper(self._client)
        )
        if ncall is None:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
            collections,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collections: list[SQLiteCollection],
        helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if len(collections) == 1:
            op_converter = collections[0].op_converter
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            table = self._get_table_name(op_parser)
            id_column = ParameterParser.get_collection_parameter(
                self.id_column, table
            )
            value_column = ParameterParser.get_collection_parameter(
                self.value_column, table
            )
            pk_column = ParameterParser.get_collection_parameter(
                self.pk_column, table
            )
            args = {
                "table": table,
                "config": get_collection_config(op_parser),
                "columns": {
                    "id_column": id_column,
                    "value_column": value_column,
                    "pk_column": pk_column,
                },
                "exists": op_parser.get_where_exists(),
            }
            call = NCall(
                helper.create_collection,
                args,
            )
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "table": self._get_table_name(op_parser),
                "exists": op_parser.get_where_exists(),
            }
            call = NCall(
                helper.drop_collection,
                args,
            )
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = OperationConverter.convert_list_collections()
            call = NCall(helper.execute, args)
        # HAS COLLECTION
        elif op_parser.op_equals(StoreOperation.HAS_COLLECTION):
            args = OperationConverter.convert_has_collection(
                self._get_table_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # CREATE INDEX
        elif op_parser.op_equals(StoreOperation.CREATE_INDEX):
            table = self._get_table_name(op_parser)
            value_column = ParameterParser.get_collection_parameter(
                self.value_column, table
            )
            args = {
                "index": op_parser.get_index(),
                "table": table,
                "value_column": value_column,
                "exists": op_parser.get_where_exists(),
            }
            call = NCall(
                helper.create_index,
                args,
            )
        # DROP INDEX
        elif op_parser.op_equals(StoreOperation.DROP_INDEX):
            args = {
                "index": op_parser.get_index(),
                "table": self._get_table_name(op_parser),
                "exists": op_parser.get_where_exists(),
            }
            call = NCall(
                helper.drop_index,
                args,
            )
        # LIST INDEXES
        elif op_parser.op_equals(StoreOperation.LIST_INDEXES):
            args = {
                "table": self._get_table_name(op_parser),
            }
            call = NCall(helper.list_indexes, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = op_converter.convert_get(op_parser.get_key())
            call = NCall(helper.execute, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, state = op_converter.convert_put(
                op_parser.get_key(),
                op_parser.get_value(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: PreconditionFailedError},
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            args, state = op_converter.convert_update(
                op_parser.get_key(),
                op_parser.get_set(),
                op_parser.get_where(),
                op_parser.get_returning_as_bool(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: PreconditionFailedError},
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = op_converter.convert_delete(
                op_parser.get_key(), op_parser.get_where()
            )
            call = NCall(helper.execute, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = op_converter.convert_query(
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            call = NCall(helper.execute, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = op_converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(helper.execute, args)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, state = op_converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            call = NCall(helper.batch, args, None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            args, state = OperationConverter.convert_transact(
                op_parser.get_operation_parsers(),
                [col.op_converter for col in collections],
            )
            call = NCall(helper.transact, args, None)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collections: list[SQLiteCollection],
    ) -> Any:
        if len(collections) == 1:
            result_converter = collections[0].result_converter
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = nresult
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = nresult
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = ResultConverter.convert_list_collections(nresult)
        # HAS COLLECTION
        elif op_parser.op_equals(StoreOperation.HAS_COLLECTION):
            result = ResultConverter.convert_has_collection(nresult)
        # CREATE INDEX
        if op_parser.op_equals(StoreOperation.CREATE_INDEX):
            result = nresult
        # DROP INDEX
        elif op_parser.op_equals(StoreOperation.DROP_INDEX):
            result = nresult
        # LIST INDEXES
        elif op_parser.op_equals(StoreOperation.LIST_INDEXES):
            result = nresult
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            result = result_converter.convert_get(nresult)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = result_converter.convert_put(nresult, op_parser, state)
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            result = result_converter.convert_update(nresult, op_parser, state)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = result_converter.convert_delete(nresult, op_parser)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            result = result_converter.convert_query(nresult, op_parser)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = result_converter.convert_count(nresult)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = result_converter.convert_batch(nresult, state)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            result = []
            op_parsers = op_parser.get_operation_parsers()
            for i in range(0, len(op_parsers)):
                if state is not None:
                    c = state["values"][i]
                result.append(
                    self._convert_nresult(
                        nresult[i], c, op_parsers[i], [collections[i]]
                    )
                )
        return result


class Helper:
    @staticmethod
    def normalize_select(
        processor: ItemProcessor, select: Select, item: tuple
    ) -> dict:
        result: dict = {}
        for i in range(0, len(select.terms)):
            DataAccessor.update_field(
                result,
                processor.resolve_field(select.terms[i].field),
                UpdateOp.PUT,
                json.loads(item[i]) if item[i] is not None else None,
            )
        return result


class ResultConverter:
    processor: ItemProcessor

    def __init__(
        self,
        processor: ItemProcessor,
    ):
        self.processor = processor

    @staticmethod
    def convert_list_collections(nresult: Any):
        result = []
        for item in nresult:
            result.append(item[0])
        return result

    @staticmethod
    def convert_has_collection(nresult: Any):
        return nresult is not None

    @staticmethod
    def convert_list_indexes(nresult: Any, use_name_type: bool):
        def match_brackets(text):
            stack = []
            results = []
            start_idx = -1

            for i, char in enumerate(text):
                if char == "(":
                    if not stack:
                        start_idx = i
                    stack.append(i)
                elif char == ")":
                    stack.pop()
                    if not stack:
                        start = start_idx + 1
                        results.append(text[start:i])
            return results

        def extract_fields(definition: str) -> list[dict]:
            matches = match_brackets(definition)
            match = matches[0]
            path_type_pattern = r"->\s*'([^']+)'|\->>\s*'([^']+)'|\)\s*"
            order_pattern = r"(DESC|ASC)$"
            splits = str(match).split(",")
            results: list[dict] = []
            for field_expr in splits:
                field_parts = []
                type_ = "text"

                # Extract -> and ->>
                for path_match in re.finditer(path_type_pattern, field_expr):
                    if path_match.group(1):  # Matches -> 'field'
                        field_parts.append(path_match.group(1))
                    elif path_match.group(2):  # Matches ->> 'field'
                        field_parts.append(path_match.group(2))

                # Extract order (DESC/ASC)
                order_match = re.search(order_pattern, field_expr)
                order_modifier = order_match.group(1) if order_match else "ASC"

                # Combine field parts into dot-separated path
                field = ".".join(field_parts)

                # Append the result
                results.append(
                    {"field": field, "type": type_, "order": order_modifier}
                )
            return results

        result: list = []
        for row in nresult:
            index_name, index_def = row
            if "sqlite_autoindex" in index_name:
                continue
            if "," not in index_def:
                field_def = extract_fields(index_def)[0]
                field = field_def["field"]
                type_from_name = BaseIndexHelper.get_type_from_name(index_name)
                if "DESC" in index_def:
                    result.append(DescIndex(name=index_name, field=field))
                elif "ASC" in index_def:
                    result.append(AscIndex(name=index_name, field=field))
                elif use_name_type and type_from_name == "array":
                    result.append(ArrayIndex(name=index_name, field=field))
                elif use_name_type and type_from_name == "hash":
                    result.append(HashIndex(name=index_name, field=field))
                elif use_name_type and type_from_name == "range":
                    result.append(RangeIndex(name=index_name, field=field))
                elif use_name_type and type_from_name == "field":
                    result.append(FieldIndex(name=index_name, field=field))
                else:
                    result.append(
                        AscIndex(
                            name=index_name,
                            field=field_def["field"],
                        )
                    )
            elif "," in index_def:
                index_parts: list = []
                field_defs = extract_fields(index_def)
                name_types: list = (
                    BaseIndexHelper.get_composite_types_from_name(index_name)
                )
                if len(field_defs) != len(name_types):
                    name_types = [None] * len(field_defs)
                for field_def, name_type in zip(field_defs, name_types):
                    field = field_def["field"]
                    if field_def["order"] == "DESC":
                        index_parts.append(DescIndex(field=field))
                    else:
                        if use_name_type and name_type == "array":
                            index_parts.append(ArrayIndex(field=field))
                        elif use_name_type and name_type == "range":
                            index_parts.append(RangeIndex(field=field))
                        elif use_name_type and name_type == "hash":
                            index_parts.append(HashIndex(field=field))
                        elif use_name_type and name_type == "field":
                            index_parts.append(FieldIndex(field=field))
                        else:
                            index_parts.append(AscIndex(field=field))
                result.append(
                    CompositeIndex(name=index_name, fields=index_parts)
                )
            else:
                result.append(FieldIndex(name=index_name, field=index_def))
        return result

    def convert_get(self, nresult: Any):
        if nresult is None:
            raise NotFoundError
        return build_item_from_value(
            processor=self.processor,
            value=json.loads(nresult[0]),
            include_value=True,
        )

    def convert_put(
        self, nresult: Any, op_parser: StoreOperationParser, state: dict | None
    ):
        where = op_parser.get_where()
        exists = op_parser.get_where_exists()
        returning = op_parser.get_returning_as_bool()
        if exists is False:
            pass
        elif where is not None:
            if nresult == 0:
                raise PreconditionFailedError
        value: Any = state["value"] if state is not None else None
        return build_item_from_value(
            processor=self.processor,
            value=value,
            include_value=(returning or False),
        )

    def convert_update(
        self, nresult: Any, op_parser: StoreOperationParser, state: dict | None
    ):
        where = op_parser.get_where()
        returning = op_parser.get_returning_as_bool()
        if where is None:
            if returning is None:
                if nresult == 0:
                    raise NotFoundError
            else:
                if nresult is None or len(nresult) == 0:
                    raise NotFoundError
        elif where is not None:
            if returning is None:
                if nresult == 0:
                    raise PreconditionFailedError
            else:
                if nresult is None or len(nresult) == 0:
                    raise PreconditionFailedError
        if isinstance(nresult, tuple):
            result = build_item_from_value(
                processor=self.processor,
                value=json.loads(nresult[0]),
                include_value=True,
            )
        else:
            key = op_parser.get_key()
            result = build_item_from_parts(
                processor=self.processor,
                key=key,
                etag=(
                    state["etag"]
                    if state is not None and "etag" in state
                    else None
                ),
            )
        return result

    def convert_delete(self, nresult: Any, op_parser: StoreOperationParser):
        where = op_parser.get_where()
        if where is None:
            if nresult == 0:
                raise NotFoundError
        elif where is not None:
            if nresult == 0:
                raise PreconditionFailedError
        return None

    def convert_query(self, nresult: Any, op_parser: StoreOperationParser):
        items: list = []
        select = op_parser.get_select()
        for item in nresult:
            if select is None or len(select.terms) == 0:
                nvalue = json.loads(item[0])
            else:
                nvalue = Helper.normalize_select(self.processor, select, item)
            items.append(
                build_item_from_value(
                    processor=self.processor, value=nvalue, include_value=True
                )
            )
        return build_query_result(items)

    def convert_count(self, nresult: Any):
        return nresult[0]

    def convert_batch(self, nresult: Any, state: dict | None):
        result: list = []
        if state is not None:
            for c in state["values"]:
                if c is not None:
                    result.append(
                        build_item_from_value(
                            processor=self.processor,
                            value=c["value"],
                        )
                    )
                else:
                    result.append(None)
        return result


class OperationConverter:
    processor: ItemProcessor
    table: str
    id_column: str
    value_column: str
    pk_column: str | None

    FIELD_TYPE_TEXT: str = "TEXT"
    FIELD_TYPE_NUMERIC: str = "NUMERIC"
    FIELD_TYPE_BOOLEAN: str = "BOOLEAN"
    FIELD_TYPE_JSONB: str = "JSONB"

    def __init__(
        self,
        processor: ItemProcessor,
        table: str,
        id_column: str,
        value_column: str,
        pk_column: str | None,
    ):
        self.processor = processor
        self.table = table
        self.id_column = id_column
        self.value_column = value_column
        self.pk_column = pk_column

    @staticmethod
    def convert_create_collection(
        table: str,
        config: DocumentCollectionConfig | None,
        columns: dict,
        exists: bool | None,
    ) -> dict:
        def get_column_type(type: str) -> str:
            if type == DocumentFieldType.STRING:
                return "TEXT"
            elif type == DocumentFieldType.NUMBER:
                return "NUMERIC"
            raise BadRequestError("Field type not supported")

        pk_column = (
            config.pk_field
            if config is not None and config.pk_field is not None
            else columns["pk_column"]
        )
        pk_type = get_column_type(
            config.pk_type
            if config is not None and config.pk_type is not None
            else DocumentFieldType.STRING
        )
        id_column = (
            config.id_field
            if config is not None and config.id_field is not None
            else columns["id_column"]
        )
        id_type = get_column_type(
            config.id_type
            if config is not None and config.id_type is not None
            else DocumentFieldType.STRING
        )
        value_column = (
            config.value_field
            if config is not None and config.value_field is not None
            else columns["value_column"]
        )
        if exists is False:
            if_not_exists_clause = ""
        else:
            # Handle exceptions upstream.
            # if_not_exists_clause = " IF NOT EXISTS"
            if_not_exists_clause = ""
        if pk_column is not None:
            query = f"""CREATE TABLE{if_not_exists_clause} {table}
                ({id_column} {id_type} PRIMARY KEY,
                {pk_column} {pk_type},
                {value_column} JSON,
                UNIQUE({id_column}))
                """
        else:
            query = f"""CREATE TABLE{if_not_exists_clause} {table}
                ({id_column} {id_type} PRIMARY KEY,
                {value_column} JSON,
                UNIQUE({id_column}))
                """
        return {"query": query}

    @staticmethod
    def convert_drop_collection(table: str, exists: bool | None) -> dict:
        if exists is True:
            if_exists = ""
        else:
            # Handle exception upstream.
            # if_exists = " IF EXISTS"
            if_exists = ""
        str = f"DROP TABLE{if_exists} {table}"
        return {"query": str}

    @staticmethod
    def convert_list_collections() -> dict:
        str = "SELECT name FROM sqlite_master WHERE type='table'"
        return {"query": str, "fetchall": True}

    @staticmethod
    def convert_has_collection(table: str) -> dict:
        str = f"""
            SELECT name FROM sqlite_master
            WHERE type='table' AND name='{table}'
            """
        return {"query": str}

    @staticmethod
    def convert_create_index(
        index: Index,
        table: str,
        value_column: str,
        exists: bool | None,
    ) -> dict | None:
        def get_field_type(field_type) -> str | None:
            field_type_map = {
                DocumentFieldType.STRING.value: "TEXT",
                DocumentFieldType.NUMBER.value: "NUMERIC",
                DocumentFieldType.BOOLEAN.value: "BOOLEAN",
            }
            if field_type in field_type_map:
                return field_type_map[field_type]
            return None

        if exists is False:
            if_not_exists = ""
        else:
            # Handle exception upstream.
            # if_not_exists = " IF NOT EXISTS"
            if_not_exists = ""
        queries: list = []
        name = BaseIndexHelper.convert_index_name(index, collection=table)
        create_str = f"CREATE INDEX{if_not_exists} {name} ON {table} "
        if isinstance(
            index,
            (
                RangeIndex,
                FieldIndex,
                HashIndex,
            ),
        ):
            field_type = get_field_type(index.field_type)
            path = OperationConverter._convert_field(
                value_column,
                index.field,
                field_type,
            )
            query = f"{create_str} (({path}))"
            queries.append({"query": query})
        elif isinstance(index, AscIndex):
            path = OperationConverter._convert_field(
                value_column,
                index.field,
                get_field_type(index.field_type),
            )
            query = f"{create_str} (({path}) ASC)"
            queries.append({"query": query})
        elif isinstance(index, DescIndex):
            path = OperationConverter._convert_field(
                value_column,
                index.field,
                get_field_type(index.field_type),
            )
            query = f"{create_str} (({path}) DESC)"
            queries.append({"query": query})
        elif isinstance(index, CompositeIndex):
            index_str_parts: list = []
            for part in index.fields:
                path = OperationConverter._convert_field(
                    value_column,
                    part.field,
                    get_field_type(part.field_type),
                )
                if isinstance(part, DescIndex):
                    index_str_parts.append(f"({path}) DESC")
                elif isinstance(
                    part, (AscIndex, FieldIndex, RangeIndex, HashIndex)
                ):
                    index_str_parts.append(f"({path}) ASC")
                else:
                    # Not supported
                    return None
            index_str = ", ".join(index_str_parts)
            query = f"{create_str} ({index_str})"
            queries.append({"query": query})
        else:
            return None
        return {"ops": queries}

    @staticmethod
    def convert_drop_index(
        index: Index, table: str, exists: bool | None
    ) -> dict:
        if exists is True:
            if_exists = ""
        else:
            # Handle exception upstream.
            # if_exists = " IF EXISTS"
            if_exists = ""
        queries: list = []
        name = BaseIndexHelper.convert_index_name(index, collection=table)
        query = f"DROP INDEX{if_exists} {name}"
        queries.append({"query": query})
        return {"ops": queries}

    @staticmethod
    def convert_list_indexes(table: str) -> dict:
        query = f"""
        SELECT name, sql
        FROM sqlite_master
        WHERE type = 'index' AND tbl_name = '{table}'
        """
        return {"query": query, "fetchall": True}

    @staticmethod
    def convert_transact(
        op_parsers: list[StoreOperationParser],
        converters: list[OperationConverter],
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for i in range(0, len(op_parsers)):
            op_parser = op_parsers[i]
            converter = converters[i]
            if op_parser.op_equals(StoreOperation.PUT):
                args, state = converter.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(args)
                states.append(state)
            elif op_parser.op_equals(StoreOperation.UPDATE):
                args, state = converter.convert_update(
                    op_parser.get_key(),
                    op_parser.get_set(),
                    op_parser.get_where(),
                    op_parser.get_returning_as_bool(),
                )
                ops.append(args)
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = converter.convert_delete(
                    op_parser.get_key(), op_parser.get_where()
                )
                ops.append(args)
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.PUT):
                args, state = self.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(args)
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = self.convert_delete(
                    op_parser.get_key(), op_parser.get_where()
                )
                ops.append(args)
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_get(self, key: Value) -> dict:
        str = f"""SELECT {self.value_column} FROM {self.table}
                    WHERE {self.convert_key_comparison(key)}"""
        return {"query": str}

    def convert_put(
        self,
        key: Value,
        value: dict,
        where: Expression | None,
        exists: bool | None,
    ) -> tuple[dict, dict | None]:
        document = self.processor.add_embed_fields(value, key)
        if key is not None:
            key = self.processor.get_key_from_key(key)
        else:
            key = self.processor.get_key_from_value(document)
        if (where is None and exists is None) or exists is False:
            id, pk = self.processor.get_id_pk_from_key(key)
            if self.pk_column is not None:
                columns = f"""({self.id_column},
                    {self.pk_column}, {self.value_column})"""
                json_string = self._safe_json_dumps(document)
                values = f"""({self.convert_expr(id)},
                    {self.convert_expr(pk)}, '{json_string}')"""
            else:
                columns = f"({self.id_column}, {self.value_column})"
                json_string = self._safe_json_dumps(document)
                values = f"({self.convert_expr(id)}, '{json_string}')"
            query = f"INSERT INTO {self.table} {columns} VALUES {values}"
            if exists is None:
                query = f"""{query} ON CONFLICT ({self.id_column})
                    DO UPDATE SET {self.value_column} =
                    EXCLUDED.{self.value_column}"""
        elif exists is True or where is not None:
            where_expr = self.convert_key_comparison(key)
            if exists is not True and where is not None:
                where_expr = f"{where_expr} AND {self.convert_expr(where)}"
            json_string = self._safe_json_dumps(document)
            query = f"""UPDATE {self.table}
                SET {self.value_column} = '{json_string}'
                WHERE {where_expr}"""
        return {"query": query, "rowcount": True}, {"value": document}

    def convert_update(
        self,
        key: Value,
        set: Update,
        where: Expression | None,
        returning: bool | None,
    ) -> tuple[dict, dict | None]:
        uset = copy.deepcopy(set)
        state = None
        if self.processor.needs_local_etag():
            etag = self.processor.generate_etag()
            uset = self.processor.add_etag_update(uset, etag)
            state = {"etag": etag}
        str = f"UPDATE {self.table} SET {self.convert_update_ops(uset)}"
        where_expr = self.convert_key_comparison(key)
        if where is not None:
            where_expr = f"{where_expr} AND {self.convert_expr(where)}"
        str = f"{str} WHERE {where_expr}"
        if returning:
            str = f"{str} RETURNING {self.value_column}"
        if not returning:
            return {"query": str, "rowcount": True}, state
        return {"query": str}, state

    def convert_delete(self, key: Value, where: Expression | None) -> dict:
        where_expr = self.convert_key_comparison(key)
        if where is not None:
            where_expr = f"{where_expr} AND {self.convert_expr(where)}"
        str = f"DELETE FROM {self.table} WHERE {where_expr}"
        return {"query": str, "rowcount": True}

    def convert_query(
        self,
        select: Select | None = None,
        where: Expression | None = None,
        order_by: OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        str = f"""SELECT {self.convert_select(
            select)} FROM {self.table}"""
        if where is not None:
            str = f"{str} WHERE {self.convert_expr(where)}"
        if order_by is not None:
            str = f"""{str} ORDER BY {self.convert_order_by(
                order_by)}"""
        if limit is not None:
            if offset is not None:
                str = f"""{str} LIMIT {limit} OFFSET {offset}"""
            else:
                str = f"""{str} LIMIT {limit}"""
        return {"query": str, "fetchall": True}

    def convert_count(self, where: Expression | None = None) -> dict:
        str = f"SELECT COUNT(*) FROM {self.table}"
        if where is not None:
            str = f"{str} WHERE {self.convert_expr(where)}"
        return {"query": str}

    def convert_key_comparison(self, key: Value) -> str:
        id, pk = self.processor.get_id_pk_from_key(key)
        expr = f"{self.id_column} = {self.convert_expr(id)}"
        if self.pk_column is not None:
            expr = f"{expr} AND {self.pk_column} = {self.convert_expr(pk)}"
        return expr

    @staticmethod
    def _convert_field(
        value_column: str,
        field: str,
        field_type: str | None = None,
    ) -> str:
        path = (
            field.replace(".", "/")
            .replace("[", "/")
            .replace("]", "")
            .rstrip("/")
        )
        splits = path.split("/")
        nsplits = []
        for split in splits:
            if split.isnumeric():
                nsplits.append(split)
            else:
                nsplits.append(f"'{split}'")
        splits = nsplits
        splits.insert(0, value_column)
        if (
            field_type == OperationConverter.FIELD_TYPE_TEXT
            or field_type == OperationConverter.FIELD_TYPE_NUMERIC
            or field_type == OperationConverter.FIELD_TYPE_BOOLEAN
        ):
            path = "->".join(splits[0:-1])
            path = f"{path}->>{splits[-1]}"
        else:
            path = "->".join(splits)
        return path

    def convert_field(
        self,
        field: Field | str,
        field_type: str | None = None,
    ) -> str:
        path: str
        if isinstance(field, str):
            path = self.processor.resolve_field(field)
        elif isinstance(field, Field):
            path = self.processor.resolve_field(field.path)
        return OperationConverter._convert_field(
            self.value_column,
            path,
            field_type,
        )

    def convert_expr(self, expr: Expression | None) -> str:
        if expr is None:
            return "NULL"
        if isinstance(expr, str):
            return f"'{expr}'"
        if isinstance(expr, bool):
            return str(expr).upper()
        if isinstance(expr, (int, float)):
            return json.dumps(expr)
        if isinstance(expr, (dict, list)):
            json_string = self._safe_json_dumps(expr, separators=(",", ":"))
            return f"'{json_string}'"
        if isinstance(expr, Field):
            return self.convert_field(expr)
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            return self.convert_comparison(expr)
        if isinstance(expr, And):
            return f"""({self.convert_expr(
                expr.lexpr)} AND {self.convert_expr(
                    expr.rexpr)})"""
        if isinstance(expr, Or):
            return f"""({self.convert_expr(
                expr.lexpr)} OR {self.convert_expr(
                    expr.rexpr)})"""
        if isinstance(expr, Not):
            return f"""NOT {self.convert_expr(
                expr.expr)}"""
        return str(expr)

    def convert_comparison(self, expr: Comparison) -> str:
        def get_field_type(value: Any) -> str | None:
            ftype = None
            if isinstance(value, bool):
                ftype = OperationConverter.FIELD_TYPE_BOOLEAN
            elif isinstance(value, (int, float)):
                ftype = OperationConverter.FIELD_TYPE_NUMERIC
            elif isinstance(value, str):
                ftype = OperationConverter.FIELD_TYPE_TEXT
            elif isinstance(value, list) and len(value) > 0:
                return get_field_type(value[0])
            return ftype

        value = None
        if isinstance(expr.lexpr, Field):
            value = expr.rexpr
        elif isinstance(expr.rexpr, Field):
            value = expr.lexpr
        field_type = get_field_type(value)

        if isinstance(expr.lexpr, Field):
            lhs = self.convert_field(expr.lexpr, field_type)
        else:
            lhs = self.convert_expr(expr.lexpr)
        if isinstance(expr.rexpr, Field):
            rhs = self.convert_field(expr.rexpr, field_type)
        else:
            rhs = self.convert_expr(expr.rexpr)

        if expr.op == ComparisonOp.BETWEEN and isinstance(expr.rexpr, list):
            return f"""{lhs} >= {self.convert_expr(
                expr.rexpr[0])} AND {lhs} <= {self.convert_expr(
                    expr.rexpr[1])}"""
        elif (
            expr.op == ComparisonOp.IN or expr.op == ComparisonOp.NIN
        ) and isinstance(expr.rexpr, list):
            lst = ", ".join(self.convert_expr(i) for i in expr.rexpr)
            rhs = f"({lst})"
        return f"""{lhs} {expr.op.value} {rhs}"""

    def convert_func(self, expr: Function) -> str:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == QueryFunctionName.IS_TYPE:
                field = self.processor.resolve_field(args[0])
                field = f"$.{field}"
                field_type = DocumentFieldType(args[1].lower())
                if field_type == DocumentFieldType.OBJECT:
                    field_types = ["'object'"]
                elif field_type == DocumentFieldType.ARRAY:
                    field_types = ["'array'"]
                elif field_type == DocumentFieldType.BOOLEAN:
                    field_types = ["'false'", "'true'"]
                elif field_type == DocumentFieldType.NUMBER:
                    field_types = ["'integer'", "'real'"]
                elif field_type == DocumentFieldType.STRING:
                    field_types = ["'text'"]
                elif field_type == DocumentFieldType.NULL:
                    field_types = ["'null'"]
                in_arr = f"({','.join(field_types)})"
                return f"JSON_TYPE({self.value_column}, '{field}') IN {in_arr}"
            if name == QueryFunctionName.IS_DEFINED:
                field = self.convert_field(args[0])
                return f"{field} IS NOT NULL"
            if name == QueryFunctionName.IS_NOT_DEFINED:
                field = self.convert_field(args[0])
                return f"{field} IS NULL"
            if name == QueryFunctionName.LENGTH:
                field = self.convert_field(
                    args[0], OperationConverter.FIELD_TYPE_TEXT
                )
                return f"LENGTH({field})"
            if name == QueryFunctionName.CONTAINS:
                field = self.convert_field(
                    args[0], OperationConverter.FIELD_TYPE_TEXT
                )
                value = args[1]
                return f"{field} LIKE '%{value}%'"
            if name == QueryFunctionName.STARTS_WITH:
                field = self.convert_field(
                    args[0], OperationConverter.FIELD_TYPE_TEXT
                )
                value = args[1]
                return f"{field} LIKE '{value}%'"
            if name == QueryFunctionName.ARRAY_LENGTH:
                field = self.convert_field(args[0])
                return f"JSON_ARRAY_LENGTH({field})"
            if name == QueryFunctionName.ARRAY_CONTAINS:
                field = self.processor.resolve_field(args[0])
                field = f"$.{field}"
                value = self.convert_expr(args[1])
                where = f"value = {value}"
                json_each = (
                    f"JSON_EACH({self.table}.{self.value_column}, '{field}')"
                )
                query = f"SELECT 1 FROM {json_each} WHERE {where}"
                query = f"EXISTS ({query})"
                return query
            if name == QueryFunctionName.ARRAY_CONTAINS_ANY:
                field = self.processor.resolve_field(args[0])
                field = f"$.{field}"
                clauses = []
                for item in args[1]:
                    value = self.convert_expr(item)
                    clauses.append(f"value = {value}")
                where = f"({str.join(' OR ', clauses)})"
                json_each = (
                    f"JSON_EACH({self.table}.{self.value_column}, '{field}')"
                )
                query = f"SELECT 1 FROM {json_each} WHERE {where}"
                query = f"EXISTS ({query})"
                return query
        raise BadRequestError(f"Function {name} not supported")

    def convert_order_by(self, order_by: OrderBy) -> str:
        str_terms = []
        for term in order_by.terms:
            _str_term = self.convert_field(term.field)
            if term.direction is not None:
                _str_term = f"{_str_term} {term.direction.value}"
            str_terms.append(_str_term)
        return ", ".join([t for t in str_terms])

    def convert_select(self, select: Select | None) -> str:
        if select is None or len(select.terms) == 0:
            return self.value_column
        str_terms = []
        for term in select.terms:
            _str_term = self.convert_expr(Field(path=term.field))
            if term.alias is not None:
                _str_term = f"{_str_term} AS {term.alias}"
            str_terms.append(_str_term)
        return ", ".join([t for t in str_terms])

    def convert_update_ops(self, update: Update) -> str:
        def get_field_path(field: str) -> tuple:
            field = self.processor.resolve_field(field)
            path = (
                field.replace(".", "/")
                .replace("[", "/")
                .replace("]", "")
                .rstrip("/")
            )
            splits = path.split("/")
            if field.endswith("[-]"):
                field = field.replace("[-]", "[#]")
            field = f"'$.{field}'"
            return field, splits

        str = self.value_column
        for operation in update.operations:
            op = operation.op
            field_path, splits = get_field_path(operation.field)
            if op == UpdateOp.PUT:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0], (dict, list)):
                    value = f"json({value})"
                str = f"JSON_SET({str}, {field_path}, {value})"
            elif op == UpdateOp.INSERT:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0], (dict, list)):
                    value = f"json({value})"
                if splits[-1] == "-":
                    str = f"JSON_INSERT({str}, {field_path}, {value})"
                elif splits[-1].isnumeric():
                    index = int(splits[-1])
                    array_field = f"{field_path[: field_path.rfind('[')]}'"
                    if isinstance(operation.args[0], (dict, list)):
                        group = "JSON_GROUP_ARRAY(json(t5.value))"
                    else:
                        group = "JSON_GROUP_ARRAY(t5.value)"
                    value_query = f"""(SELECT {group}
                        FROM (SELECT t2.value
                        FROM (SELECT t1.value
                        FROM JSON_EACH({self.table}.{self.value_column},
                        {array_field})
                        AS t1 LIMIT {index})
                        AS t2
                        UNION ALL
                        SELECT {value}
                        UNION ALL
                        SELECT t4.value
                        FROM (SELECT t3.value
                        FROM JSON_EACH({self.table}.{self.value_column},
                        {array_field})
                        AS t3 LIMIT -1 OFFSET {index})
                        AS t4)
                        AS t5)"""
                    str = f"JSON_SET({str}, {array_field}, {value_query})"
                else:
                    str = f"JSON_SET({str}, {field_path}, {value})"
            elif op == UpdateOp.DELETE:
                str = f"JSON_REMOVE({str}, {field_path})"
            elif op == UpdateOp.INCREMENT:
                value = operation.args[0]
                extract = f"JSON_EXTRACT({self.value_column}, {field_path})"
                extract = f"IFNULL({extract}, 0)"
                extract = f"{extract} + {value}"
                str = f"JSON_SET({str}, {field_path}, {extract})"
            elif op == UpdateOp.MOVE:
                dest_field_path, s = get_field_path(operation.args[0].path)
                extract = f"JSON_EXTRACT({self.value_column}, {field_path})"
                str = f"JSON_SET({str}, {dest_field_path}, {extract})"
                str = f"JSON_REMOVE({str}, {field_path})"
            elif op == UpdateOp.ARRAY_UNION:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0][0], (dict, list)):
                    group = "JSON_GROUP_ARRAY(json(t3.value))"
                else:
                    group = "JSON_GROUP_ARRAY(t3.value)"
                value_query = f"""(SELECT {group}
                    FROM (SELECT t1.value
                    FROM JSON_EACH({self.table}.{self.value_column},
                    {field_path})
                    AS t1
                    UNION ALL
                    SELECT t2.value
                    FROM JSON_EACH({value})
                    AS t2
                    WHERE NOT EXISTS (
                    SELECT 1
                    FROM JSON_EACH({self.table}.{self.value_column},
                    {field_path})
                    WHERE value = t2.value))
                    AS t3)
                    """
                str = f"JSON_SET({str}, {field_path}, {value_query})"
            elif op == UpdateOp.ARRAY_REMOVE:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0][0], (dict, list)):
                    group = "JSON_GROUP_ARRAY(json(t2.value))"
                else:
                    group = "JSON_GROUP_ARRAY(t2.value)"
                value_query = f"""(SELECT {group}
                    FROM (SELECT t1.value
                    FROM JSON_EACH({self.table}.{self.value_column},
                    {field_path})
                    AS t1
                    WHERE NOT EXISTS (
                    SELECT 1
                    FROM JSON_EACH({value})
                    WHERE value = t1.value))
                    AS t2)
                    """
                str = f"JSON_SET({str}, {field_path}, {value_query})"
            else:
                raise BadRequestError(
                    f"Operation {operation.op.value} is not supported"
                )
        return f"{self.value_column} = {str}"

    def _safe_json_dumps(self, value, separators=None) -> str:
        return json.dumps(value, separators=separators).replace("'", "''")


class ClientHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    def create_collection(
        self,
        table: str,
        config: DocumentCollectionConfig | None,
        columns: dict,
        exists: bool | None,
    ) -> CollectionResult:
        args: Any = OperationConverter.convert_create_collection(
            table, config, columns, exists
        )
        status = CollectionStatus.CREATED
        try:
            self.execute(**args)
        except sqlite3.OperationalError:
            status = CollectionStatus.EXISTS
            if exists is False:
                raise ConflictError

        index_results = []
        if config and config.indexes:
            for index in config.indexes:
                index_results.append(
                    self.create_index(
                        index, table, columns["value_column"], None
                    )
                )
        return CollectionResult(status=status, indexes=index_results)

    def drop_collection(
        self, table: str, exists: bool | None
    ) -> CollectionResult:
        args = OperationConverter.convert_drop_collection(table, exists)
        try:
            self.execute(**args)
        except sqlite3.OperationalError:
            if exists is True:
                raise NotFoundError
            return CollectionResult(status=CollectionStatus.NOT_EXISTS)
        return CollectionResult(status=CollectionStatus.DROPPED)

    def create_index(
        self,
        index: Index,
        table: str,
        value_column: str,
        exists: bool | None,
    ) -> IndexResult:
        indexes = self.list_indexes(table, False)
        status, match_index = BaseIndexHelper.check_index_status(
            indexes, index
        )
        if status == IndexStatus.EXISTS or status == IndexStatus.COVERED:
            if exists is False:
                raise ConflictError
            return IndexResult(status=status, index=match_index)
        args = OperationConverter.convert_create_index(
            index, table, value_column, exists
        )
        if args:
            self.batch(**args)
            return IndexResult(status=IndexStatus.CREATED)
        else:
            return IndexResult(status=IndexStatus.NOT_SUPPORTED)

    def drop_index(
        self, index: Index, table: str, exists: bool | None
    ) -> IndexResult:
        args = OperationConverter.convert_drop_index(index, table, exists)
        try:
            self.batch(**args)
        except sqlite3.OperationalError:
            if exists is True:
                raise NotFoundError
            return IndexResult(status=IndexStatus.NOT_EXISTS)
        return IndexResult(status=IndexStatus.DROPPED)

    def list_indexes(
        self,
        table: str,
        use_name_type: bool = True,
    ) -> list[Index]:
        args = OperationConverter.convert_list_indexes(table)
        nresult = self.execute(**args)
        result = ResultConverter.convert_list_indexes(nresult, use_name_type)
        return result

    def execute(
        self, query: str, rowcount: bool = False, fetchall: bool = False
    ) -> Any:
        cursor = self.client.cursor()
        try:
            cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return cursor.fetchall()
            return cursor.fetchone()
        finally:
            self.client.commit()
            cursor.close()

    def batch(self, ops: list) -> Any:
        cursor = self.client.cursor()
        try:
            for op in ops:
                cursor.execute(op["query"])
        finally:
            self.client.commit()
            cursor.close()

    def transact(self, ops: list) -> Any:
        result = []
        cursor = self.client.cursor()
        try:
            for op in ops:
                cursor.execute(op["query"])
                if "rowcount" in op and op["rowcount"]:
                    if cursor.rowcount == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(cursor.rowcount)
                else:
                    res = cursor.fetchone()
                    if res is None or len(res) == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(res)
            self.client.commit()
        finally:
            cursor.close()
        return result

    def close(self, nargs: Any) -> Any:
        self.client.close()


class AsyncClientHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    async def create_collection(
        self,
        table: str,
        config: DocumentCollectionConfig | None,
        columns: dict,
        exists: bool | None,
    ) -> CollectionResult:
        args: Any = OperationConverter.convert_create_collection(
            table, config, columns, exists
        )
        status = CollectionStatus.CREATED
        try:
            await self.execute(**args)
        except sqlite3.OperationalError:
            status = CollectionStatus.EXISTS
            if exists is False:
                raise ConflictError

        index_results = []
        if config and config.indexes:
            for index in config.indexes:
                index_results.append(
                    await self.create_index(
                        index, table, columns["value_column"], None
                    )
                )
        return CollectionResult(status=status, indexes=index_results)

    async def drop_collection(
        self, table: str, exists: bool | None
    ) -> CollectionResult:
        args = OperationConverter.convert_drop_collection(table, exists)
        try:
            await self.execute(**args)
        except sqlite3.OperationalError:
            if exists is True:
                raise NotFoundError
            return CollectionResult(status=CollectionStatus.NOT_EXISTS)
        return CollectionResult(status=CollectionStatus.DROPPED)

    async def create_index(
        self,
        index: Index,
        table: str,
        value_column: str,
        exists: bool | None,
    ) -> IndexResult:
        indexes = await self.list_indexes(table, False)
        status, match_index = BaseIndexHelper.check_index_status(
            indexes, index
        )
        if status == IndexStatus.EXISTS or status == IndexStatus.COVERED:
            if exists is False:
                raise ConflictError
            return IndexResult(status=status, index=match_index)
        args = OperationConverter.convert_create_index(
            index, table, value_column, exists
        )
        if args:
            await self.batch(**args)
            return IndexResult(status=IndexStatus.CREATED)
        else:
            return IndexResult(status=IndexStatus.NOT_SUPPORTED)

    async def drop_index(
        self, index: Index, table: str, exists: bool | None
    ) -> IndexResult:
        args = OperationConverter.convert_drop_index(index, table, exists)
        try:
            await self.batch(**args)
        except sqlite3.OperationalError:
            if exists is True:
                raise NotFoundError
            return IndexResult(status=IndexStatus.NOT_EXISTS)
        return IndexResult(status=IndexStatus.DROPPED)

    async def list_indexes(
        self,
        table: str,
        use_name_type: bool = True,
    ) -> list[Index]:
        args = OperationConverter.convert_list_indexes(table)
        nresult = await self.execute(**args)
        return ResultConverter.convert_list_indexes(nresult, use_name_type)

    async def execute(
        self, query: str, rowcount: bool = False, fetchall: bool = False
    ) -> Any:
        cursor = self.client.cursor()
        try:
            await cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return await cursor.fetchall()
            return await cursor.fetchone()
        finally:
            await self.client.commit()
            await cursor.close()

    async def batch(self, ops: list) -> Any:
        cursor = self.client.cursor()
        try:
            for op in ops:
                await cursor.execute(op["query"])
        finally:
            await self.client.commit()
            await cursor.close()

    async def transact(self, ops: list) -> Any:
        result = []
        cursor = self.client.cursor()
        try:
            for op in ops:
                await cursor.execute(op["query"])
                if "rowcount" in op and op["rowcount"]:
                    if cursor.rowcount == 0:
                        await self.client.rollback()
                        raise ConflictError
                    result.append(cursor.rowcount)
                else:
                    res = await cursor.fetchone()
                    if res is None or len(res) == 0:
                        await self.client.rollback()
                        raise ConflictError
                    result.append(res)
            await self.client.commit()
        finally:
            await cursor.close()
        return result

    async def close(self, nargs: Any) -> Any:
        await self.client.close()


class SQLiteCollection:
    op_converter: OperationConverter
    result_converter: ResultConverter
    processor: ItemProcessor

    def __init__(
        self,
        table: str,
        id_column: str,
        value_column: str,
        pk_column: str | None,
        id_map_field: str | None,
        pk_map_field: str | None,
        etag_embed_field: str | None,
        suppress_fields: list[str] | None,
    ) -> None:
        self.processor = ItemProcessor(
            etag_embed_field=etag_embed_field,
            id_map_field=id_map_field,
            pk_map_field=pk_map_field,
            local_etag=True,
            suppress_fields=suppress_fields,
        )
        self.op_converter = OperationConverter(
            self.processor,
            table,
            id_column,
            value_column,
            pk_column,
        )
        self.result_converter = ResultConverter(self.processor)



from __future__ import annotations

from x8.content.video import VideoAudio, VideoFrame
from x8.core import Component, operation

from ._models import LiveStreamInfo, LiveStreamParam


class LiveStream(Component):
    source: str | None
    min_resolution: int

    def __init__(
        self,
        source: str | None = None,
        min_resolution: int = 0,
        param: LiveStreamParam | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            source:
                Live stream source, e.g., local path or URL.
            min_resolution:
                Minimum resolution of the stream to decode.
            param:
                Live stream param.
        """
        self.source = source
        self.min_resolution = min_resolution
        if param is not None:
            self.source = param.url
        super().__init__(**kwargs)

    @operation()
    def get_info(self, **kwargs) -> LiveStreamInfo:
        """Get live stream info.

        Returns:
            Live stream info.
        """
        raise NotImplementedError

    @operation()
    def get_frame(
        self,
        **kwargs,
    ) -> VideoFrame | None:
        """Get the current frame in the live stream.

        Returns:
            Live stream frame.
        """
        raise NotImplementedError

    @operation()
    def get_audio(
        self,
        duration: float,
        format: str | None = None,
        codec: str | None = None,
        rate: int | None = None,
        channels: int | None = None,
        **kwargs,
    ) -> VideoAudio | None:
        """Get audio for a duration.

        Args:
            duration:
                Duration in seconds.
            format:
                Format of the audio segment.
                Possible values are "mp3", "wav", "ogg".
                Defaults to "mp3".
            codec:
                Codec of the audio segment.
                Possible values are "mp3", "pcm_s16le", etc.
                Defaults to "mp3".
            rate:
                Audio sample rate.
            channels:
                Number of channels. 1 is mono.

        Returns:
            Audio in live stream.
        """
        raise NotImplementedError

    @operation()
    def close(self, **kwargs) -> None:
        """Close the live stream."""
        raise NotImplementedError

    @operation()
    async def aget_info(self, **kwargs) -> LiveStreamInfo:
        """Get live stream info.

        Returns:
            Live stream info.
        """
        raise NotImplementedError

    @operation()
    async def aget_frame(
        self,
        **kwargs,
    ) -> VideoFrame | None:
        """Get the current frame in the live stream.

        Returns:
            Live stream frame.
        """
        raise NotImplementedError

    @operation()
    async def aget_audio(
        self,
        duration: float,
        format: str | None = None,
        codec: str | None = None,
        rate: int | None = None,
        channels: int | None = None,
        **kwargs,
    ) -> VideoAudio | None:
        """Get audio for a duration.

        Args:
            duration:
                Duration in seconds.
            format:
                Format of the audio segment.
                Possible values are "mp3", "wav", "ogg".
                Default to "mp3".
            codec:
                Codec of the audio segment.
                Possible values are "mp3", "pcm_s16le", etc.
                Default to "mp3".
            rate:
                Audio sample rate.
            channels:
                Number of channels. 1 is mono.

        Returns:
            Audio in live stream.
        """
        raise NotImplementedError

    @operation()
    async def aclose(self, **kwargs) -> None:
        """Close the live stream."""
        raise NotImplementedError



"""
Video processed with av.
"""

__all__ = ["AV"]

import base64
from io import BytesIO
from typing import Any

import av
import av.error
from x8.content.audio import Audio
from x8.content.image import Image
from x8.core import Context, Provider
from x8.core.exceptions import BadRequestError

from .._models import VideoAudio, VideoFrame, VideoInfo


class AV(Provider):
    _container: Any
    _video_stream: Any
    _audio_stream: Any

    def __init__(self, **kwargs):
        """Initialize."""
        self._container = None
        self._video_stream = None
        self._audio_stream = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._container is not None:
            return
        if self.__component__.source:
            open_param: Any = self.__component__.source
        elif self.__component__.content:
            if isinstance(self.__component__.content, str):
                open_param = BytesIO(
                    base64.b64decode(self.__component__.content)
                )
            else:
                open_param = BytesIO(self.__component__.content)
        elif self.__component__.stream is not None:
            open_param = self.__component__.stream
        else:
            raise BadRequestError("Video data not provided.")

        self._container = av.open(open_param)
        for stream in self._container.streams:
            if stream.type == "video":
                self._video_stream = stream
            if stream.type == "audio":
                self._audio_stream = stream

    def get_info(self, **kwargs) -> VideoInfo:
        self.__setup__()
        return VideoInfo(
            duration=self._container.duration / av.time_base,
            width=self._video_stream.width,
            height=self._video_stream.height,
            total_frames=self._video_stream.frames,
            frame_rate=self._video_stream.average_rate,
            bit_rate=self._video_stream.bit_rate,
            format=(
                self._video_stream.format.name
                if self._video_stream.format
                else None
            ),
            codec=self._video_stream.codec_context.name,
            pixel_format=self._video_stream.pix_fmt,
            start_time=(
                self._video_stream.start_time * self._video_stream.time_base
                if self._video_stream.start_time
                else None
            ),
        )

    def seek_frame(
        self,
        timestamp: float | None = None,
        frame_index: int | None = None,
        exact: bool = False,
        backward: bool = True,
        **kwargs,
    ) -> VideoFrame | None:
        self.__setup__()
        if self._video_stream is None:
            raise BadRequestError("No video stream found.")
        try:
            if exact:
                back = True
            else:
                back = backward
            if timestamp is not None:
                seek_pos = int(timestamp * av.time_base)
                self._container.seek(seek_pos, backward=back)
            elif frame_index is not None:
                target_pts = (
                    self._video_stream.time_base
                    * frame_index
                    / self._video_stream.average_rate
                )
                self._container.seek(int(target_pts), backward=back)
            else:
                raise BadRequestError(
                    "Either timestamp or frame_index must be provided."
                )

            for frame in self._container.decode(video=0):
                current_frame_timestamp = frame.pts * float(frame.time_base)
                current_frame_index = int(
                    frame.pts
                    * frame.time_base
                    * self._video_stream.average_rate
                )
                if (
                    exact
                    and timestamp is not None
                    and current_frame_timestamp < timestamp
                ):
                    continue
                elif (
                    exact
                    and frame_index is not None
                    and current_frame_index < frame_index
                ):
                    continue
                return VideoFrame(
                    timestamp=current_frame_timestamp,
                    frame_index=current_frame_index,
                    key_frame=frame.key_frame,
                    width=frame.width,
                    height=frame.height,
                    image=Image(_pil_image=frame.to_image()),
                )
        except av.error.EOFError:
            return None
        except av.error.PermissionError:
            return None
        return None

    def get_frame(
        self,
        **kwargs,
    ) -> VideoFrame | None:
        self.__setup__()
        if self._video_stream is None:
            raise BadRequestError("No video stream found.")
        try:
            for frame in self._container.decode(video=0):
                frame_timestamp = frame.pts * float(frame.time_base)
                frame_index = int(
                    frame.pts
                    * frame.time_base
                    * self._video_stream.average_rate
                )
                return VideoFrame(
                    timestamp=frame_timestamp,
                    frame_index=frame_index,
                    key_frame=frame.key_frame,
                    width=frame.width,
                    height=frame.height,
                    image=Image(_pil_image=frame.to_image()),
                )
            return None
        except av.error.EOFError:
            return None

    def get_audio(
        self,
        start: float | None = None,
        end: float | None = None,
        format: str | None = "mp3",
        codec: str | None = None,
        rate: int | None = None,
        channels: int | None = None,
        **kwargs,
    ) -> VideoAudio | None:
        self.__setup__()
        if self._audio_stream is None:
            raise BadRequestError("No audio stream found.")
        if format is not None:
            format = format.lower()
        output_buffer = BytesIO()
        start_pts = None
        end_pts = None
        if start is not None:
            start_pts = int(
                start
                * self._audio_stream.time_base.denominator
                / self._audio_stream.time_base.numerator
            )
            self._container.seek(int(start * av.time_base))
        if end is not None:
            end_pts = int(
                end
                * self._audio_stream.time_base.denominator
                / self._audio_stream.time_base.numerator
            )

        output_container = av.open(output_buffer, mode="w", format=format)
        if format == "wav" and codec is None:
            codec = "pcm_s16le"
        elif codec is None:
            codec = format
        args: dict = {"codec_name": codec}
        if rate is not None:
            args["rate"] = rate
        if channels == 1:
            args["layout"] = "mono"
        output_stream: Any = output_container.add_stream(**args)

        start_timestamp = None
        end_timestamp = None
        for packet in self._container.demux(self._audio_stream):
            if (
                start_pts is not None
                and packet.pts is not None
                and packet.pts < start_pts
            ):
                continue
            if (
                end_pts is not None
                and packet.pts is not None
                and packet.pts > end_pts
            ):
                break
            for frame in packet.decode():
                current_frame_timestamp = frame.pts * float(frame.time_base)
                if start_timestamp is None:
                    start_timestamp = current_frame_timestamp
                end_timestamp = current_frame_timestamp
                for out_packet in output_stream.encode(frame):
                    output_container.mux(out_packet)

        # Flush the encoder and close the output container
        for out_packet in output_stream.encode(None):
            output_container.mux(out_packet)
        output_container.close()

        output_buffer.seek(0)
        if start_timestamp is None and end_timestamp is None:
            return None

        return VideoAudio(
            start=start_timestamp,
            end=end_timestamp,
            format=format,
            codec=codec,
            audio=Audio(stream=output_buffer),
        )

    def close(self) -> None:
        if self._container is not None:
            self._container.close()
            self._container = None

    def save(
        self,
        path: str,
    ) -> None:
        if self.__component__.content:
            if isinstance(self.__component__.content, str):
                with open(path, "wb") as f:
                    f.write(base64.b64decode(self.__component__.content))
            else:
                with open(path, "wb") as f:
                    f.write(self.__component__.content)
        elif self.__component__.source:
            source = self.__component__.source
            if source.lower().startswith(
                "http://"
            ) or source.lower().startswith("https://"):
                import httpx

                response = httpx.get(self.__component__.source)
                response.raise_for_status()
                with open(path, "wb") as f:
                    f.write(response.content)
            else:
                with open(source, "rb") as src_file:
                    with open(path, "wb") as dst_file:
                        dst_file.write(src_file.read())
        elif self.__component__.stream:
            self.__component__.stream.seek(0)
            with open(path, "wb") as f:
                f.write(self.__component__.stream.read())
        else:
            raise BadRequestError("No video data to save.")



import os
import subprocess
from typing import Any, Literal

from x8._common.azure_provider import AzureProvider
from x8.compute.containerizer import DEFAULT_PLATFORM
from x8.core import Context, DataModel, Response
from x8.core.exceptions import BadRequestError, NotFoundError

from .._models import ContainerRegistryItem, ContainerRegistryItemDigest


class AzureContainerRegistryResource(DataModel):
    id: str
    name: str
    login_server: str
    location: str


class AzureContainerRegistry(AzureProvider):
    name: str | None
    platform: str
    nparams: dict[str, Any]

    _client: Any
    _aclient: Any
    _init: bool = False
    _ainit: bool = False
    _shell: bool = False

    def __init__(
        self,
        name: str | None = None,
        platform: str = DEFAULT_PLATFORM,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            name:
                Azure Container Registry name (without .azurecr.io suffix).
            credential_type:
                Type of Azure credential to use. Options are:
                - default: DefaultAzureCredential
                - client_secret: ClientSecretCredential
                - certificate: CertificateCredential
                - azure_cli: AzureCliCredential
                - shared_token_cache: SharedTokenCacheCredential
                - managed_identity: ManagedIdentityCredential
            tenant_id:
                Tenant ID for the Azure account.
            client_id:
                Client ID for the Azure account.
            client_secret:
                Client secret for the Azure account.
            certificate_path:
                Path to the certificate for the Azure account.
            platform:
                Platform to use for the container when pulling.
            nparams:
                Additional parameters for the Azure Container Registry client.
        """
        self.name = name
        self.platform = platform
        self.nparams = nparams
        self._init = False
        self._ainit = False
        self._client = None
        self._aclient = None
        self._shell = os.name == "nt"
        super().__init__(
            credential_type=credential_type,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            certificate_path=certificate_path,
            **kwargs,
        )

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        from azure.containerregistry import ContainerRegistryClient

        self._credential = self._get_credential()
        endpoint = f"https://{self._get_name()}.azurecr.io"

        self._client = ContainerRegistryClient(
            endpoint=endpoint,
            credential=self._credential,
            **self.nparams,
        )
        self._init = True

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._ainit:
            return

        from azure.containerregistry.aio import ContainerRegistryClient

        self._acredential = self._aget_credential()
        endpoint = f"https://{self._get_name()}.azurecr.io"

        self._aclient = ContainerRegistryClient(
            endpoint=endpoint,
            credential=self._acredential,
            **self.nparams,
        )
        self._ainit = True

    def _get_registry_url(self) -> str:
        return f"{self._get_name()}.azurecr.io"

    def _get_full_image_path(
        self, image_name: str, tag: str | None = None
    ) -> str:
        image = f"{self._get_registry_url()}/{image_name}"
        if tag:
            image += f":{tag}"
        return image

    def create_resource(
        self,
        resource_group: str,
        name: str | None = None,
        location: str = "westus2",
        subscription_id: str | None = None,
        sku: Literal["Basic", "Standard", "Premium"] = "Standard",
        admin_user_enabled: bool = True,
        public_network_access: Literal["Enabled", "Disabled"] = "Enabled",
        data_endpoint_enabled: bool = False,
        network_rule_bypass_options: str | None = None,
        tags: dict[str, str] | None = None,
    ) -> Response[AzureContainerRegistryResource]:
        registry_name = name or self.name

        if not registry_name:
            raise BadRequestError("Registry name is required.")
        self._ensure_resource_group(
            resource_group=resource_group,
            location=location,
            subscription_id=subscription_id,
        )

        from azure.core.exceptions import ResourceNotFoundError
        from azure.mgmt.containerregistry import (
            ContainerRegistryManagementClient,
        )
        from azure.mgmt.containerregistry.models import (
            Registry,
            RegistryUpdateParameters,
            Sku,
        )

        client = ContainerRegistryManagementClient(
            credential=self._get_credential(),
            subscription_id=subscription_id
            or self._get_default_subscription_id(),
        )
        create_params = Registry(
            location=location,
            sku=Sku(name=sku),
            admin_user_enabled=admin_user_enabled,
            public_network_access=public_network_access,
            data_endpoint_enabled=data_endpoint_enabled,
            network_rule_bypass_options=network_rule_bypass_options,
            tags=tags,
        )

        try:
            existing = client.registries.get(
                resource_group_name=resource_group,
                registry_name=registry_name,
            )
            if (existing.location or "").lower() != (location or "").lower():
                raise BadRequestError(
                    (
                        f"Registry '{registry_name}' already in location "
                        f"'{existing.location}'."
                    )
                )
            update_params = RegistryUpdateParameters(
                sku=Sku(name=sku),
                admin_user_enabled=admin_user_enabled,
                public_network_access=public_network_access,
                data_endpoint_enabled=data_endpoint_enabled,
                network_rule_bypass_options=network_rule_bypass_options,
                tags=tags,
            )
            poller = client.registries.begin_update(
                resource_group_name=resource_group,
                registry_name=registry_name,
                registry_update_parameters=update_params,
            )
            result = poller.result()
        except ResourceNotFoundError:
            poller = client.registries.begin_create(
                resource_group_name=resource_group,
                registry_name=registry_name,
                registry=create_params,
            )
            result = poller.result()

        resource = AzureContainerRegistryResource(
            id=result.id,
            name=result.name,
            login_server=result.login_server,
            location=result.location,
        )
        return Response(result=resource)

    def get_resource(
        self,
        resource_group: str,
        name: str | None = None,
        subscription_id: str | None = None,
    ) -> Response[AzureContainerRegistryResource]:
        registry_name = name or self.name
        if not registry_name:
            raise BadRequestError("Registry name is required.")

        from azure.core.exceptions import ResourceNotFoundError
        from azure.mgmt.containerregistry import (
            ContainerRegistryManagementClient,
        )

        client = ContainerRegistryManagementClient(
            credential=self._get_credential(),
            subscription_id=subscription_id
            or self._get_default_subscription_id(),
        )
        try:
            registry = client.registries.get(
                resource_group_name=resource_group,
                registry_name=registry_name,
            )
            result = AzureContainerRegistryResource(
                id=registry.id,
                name=registry.name,
                location=registry.location,
                login_server=registry.login_server,
            )
        except ResourceNotFoundError:
            raise NotFoundError("Container registry not found.")
        return Response(result=result)

    def delete_resource(
        self,
        resource_group: str,
        name: str | None = None,
        subscription_id: str | None = None,
        delete_empty_resource_group: bool = False,
    ) -> Response[None]:
        registry_name = name or self.name
        if not registry_name:
            raise BadRequestError("Registry name is required.")

        from azure.core.exceptions import ResourceNotFoundError
        from azure.mgmt.containerregistry import (
            ContainerRegistryManagementClient,
        )

        client = ContainerRegistryManagementClient(
            credential=self._get_credential(),
            subscription_id=subscription_id
            or self._get_default_subscription_id(),
        )
        try:
            poller = client.registries.begin_delete(
                resource_group_name=resource_group,
                registry_name=registry_name,
            )
            poller.wait()
        except ResourceNotFoundError:
            raise NotFoundError("Container registry not found.")
        if delete_empty_resource_group:
            self._delete_resource_group_if_empty(
                resource_group=resource_group,
                subscription_id=subscription_id,
            )
        return Response(result=None)

    def push(self, image_name: str) -> Response[ContainerRegistryItem]:
        subprocess.run(
            ["az", "acr", "login", "--name", self._get_name()],
            shell=self._shell,
            check=True,
        )

        full_path = self._get_full_image_path(image_name)
        subprocess.run(
            ["docker", "tag", image_name, full_path],
            shell=self._shell,
            check=True,
        )
        subprocess.run(
            ["docker", "push", full_path], shell=self._shell, check=True
        )

        result = ContainerRegistryItem(
            image_name=image_name, image_uri=full_path
        )
        return Response(result=result)

    def pull(
        self, image_name: str, tag: str | None = None
    ) -> Response[ContainerRegistryItem]:
        subprocess.run(
            ["az", "acr", "login", "--name", self._get_name()],
            shell=self._shell,
            check=True,
        )

        full_path = self._get_full_image_path(image_name, tag)
        subprocess.run(
            ["docker", "pull", "--platform", self.platform, full_path],
            shell=self._shell,
            check=True,
        )

        result = ContainerRegistryItem(
            image_name=image_name, image_uri=full_path
        )
        return Response(result=result)

    def tag(
        self,
        image_name: str,
        tag: str,
        digest: str | None = None,
    ) -> Response[ContainerRegistryItemDigest]:
        self.__setup__()
        if digest is None:
            latest_tag_props = self._client.get_manifest_properties(
                image_name, "latest"
            )
            digest = latest_tag_props.digest
        source_path = f"{self._get_full_image_path(image_name)}@{digest}"
        tagged_path = f"{self._get_full_image_path(image_name)}:{tag}"

        subprocess.run(
            ["docker", "pull", "--platform", self.platform, source_path],
            shell=self._shell,
            check=True,
        )
        subprocess.run(
            ["docker", "tag", source_path, tagged_path],
            shell=self._shell,
            check=True,
        )
        subprocess.run(
            ["docker", "push", tagged_path], shell=self._shell, check=True
        )

        result = ContainerRegistryItemDigest(
            image_uri=tagged_path,
            digest=digest,
            upload_time=None,
            image_size_bytes=None,
            tags=[tag],
        )
        return Response(result=result)

    def delete(
        self,
        image_name: str,
        digest: str | None = None,
        tag: str | None = None,
    ) -> Response[None]:
        self.__setup__()
        if digest is None and tag is None:
            self._client.delete_repository(image_name)
        elif digest is not None:
            self._client.delete_manifest(
                repository=image_name,
                tag_or_digest=digest,
            )
        elif tag is not None:
            self._client.delete_manifest(
                repository=image_name,
                tag_or_digest=tag,
            )
        return Response(result=None)

    def list_images(self) -> Response[list[ContainerRegistryItem]]:
        self.__setup__()
        repositories = self._client.list_repository_names()
        items = []

        for repo_name in repositories:
            items.append(
                ContainerRegistryItem(
                    image_name=repo_name,
                    image_uri=self._get_full_image_path(repo_name),
                )
            )

        result = items
        return Response(result=result)

    def get_digests(
        self, image_name: str
    ) -> Response[list[ContainerRegistryItemDigest]]:
        self.__setup__()
        manifests = self._client.list_manifest_properties(image_name)
        digests = []
        for manifest in manifests:
            upload_time = None
            if manifest.created_on:
                upload_time = manifest.created_on.timestamp()
            image_uri = (
                f"{self._get_full_image_path(image_name)}@{manifest.digest}"
            )
            digests.append(
                ContainerRegistryItemDigest(
                    image_uri=image_uri,
                    digest=manifest.digest,
                    upload_time=upload_time,
                    image_size_bytes=manifest.size_in_bytes,
                    tags=list(manifest.tags) if manifest.tags else None,
                )
            )

        result = digests
        return Response(result=result)

    async def adelete(
        self,
        image_name: str,
        digest: str | None = None,
        tag: str | None = None,
    ) -> Response[None]:
        await self.__asetup__()
        if digest is None and tag is None:
            await self._aclient.delete_repository(image_name)
        elif digest is not None:
            await self._aclient.delete_manifest(
                repository=image_name,
                tag_or_digest=digest,
            )
        elif tag is not None:
            await self._aclient.delete_manifest(
                repository=image_name,
                tag_or_digest=tag,
            )
        return Response(result=None)

    async def alist_images(self) -> Response[list[ContainerRegistryItem]]:
        await self.__asetup__()
        repositories = self._aclient.list_repository_names()
        items = []
        async for repo_name in repositories:
            items.append(
                ContainerRegistryItem(
                    image_name=repo_name,
                    image_uri=f"{self._get_registry_url()}/{repo_name}",
                )
            )

        result = items
        return Response(result=result)

    async def aget_digests(
        self, image_name: str
    ) -> Response[list[ContainerRegistryItemDigest]]:
        await self.__asetup__()
        manifests = self._aclient.list_manifest_properties(image_name)
        digests = []
        async for manifest in manifests:
            upload_time = None
            if manifest.created_on:
                upload_time = manifest.created_on.timestamp()
            image_uri = (
                f"{self._get_full_image_path(image_name)}@{manifest.digest}"
            )
            digests.append(
                ContainerRegistryItemDigest(
                    image_uri=image_uri,
                    digest=manifest.digest,
                    upload_time=upload_time,
                    image_size_bytes=manifest.size_in_bytes,
                    tags=list(manifest.tags) if manifest.tags else None,
                )
            )
        result = digests
        return Response(result=result)

    def close(self) -> Response[None]:
        self._init = False
        return Response(result=None)

    async def aclose(self) -> Response[None]:
        if self._acredential:
            await self._acredential.close()
        if self._aclient:
            await self._aclient.close()
            # Close doesn't cleanup properly in the current client
            # Might be a bug in the client, so we need the following
            await self._aclient.__aexit__(None, None, None)
        self._ainit = False
        return Response(result=None)

    def _get_name(self) -> str:
        if not self.name:
            raise BadRequestError("Registry name is required.")
        return self.name



from typing import Any

from common.secrets import get_secrets

from x8.compute.container_registry import ContainerRegistry

secrets = get_secrets()


class ContainerRegistryProvider:
    AMAZON_ELASTIC_CONTAINER_REGISTRY = "amazon_elastic_container_registry"
    AZURE_CONTAINER_REGISTRY = "azure_container_registry"
    GOOGLE_ARTIFACT_REGISTRY = "google_artifact_registry"
    DOCKER_LOCAL = "docker_local"


provider_parameters: dict[str, dict[str, Any]] = {
    ContainerRegistryProvider.AMAZON_ELASTIC_CONTAINER_REGISTRY: {
        "region": "us-west-2",
        "aws_access_key_id": secrets["aws-access-key-id"],
        "aws_secret_access_key": secrets["aws-secret-access-key"],
    },
    ContainerRegistryProvider.AZURE_CONTAINER_REGISTRY: {
        "name": secrets["azure-container-registry-name"],
    },
    ContainerRegistryProvider.GOOGLE_ARTIFACT_REGISTRY: {
        "project": secrets["google-cloud-project"],
        "location": "us-west1",
        "name": "test-registry",
    },
    ContainerRegistryProvider.DOCKER_LOCAL: {},
}


def get_component(provider_type: str):
    component = ContainerRegistry(
        __provider__=dict(
            type=provider_type,
            parameters=provider_parameters[provider_type],
        ),
    )
    return component



from typing import Literal

from x8.content.image import ImageData
from x8.content.video import VideoData
from x8.core import DataModel

VideoSize = Literal[
    "1280x720",
    "720x1280",
    "1920x1080",
    "1080x1920",
    "1792x1024",
    "1024x1792",
]


class Reference(DataModel):
    type: str | None = None
    image: ImageData | None = None


class KeyFrame(DataModel):
    time: float
    image: ImageData


class VideoGenerationResult(DataModel):
    id: str
    created_at: float | None = None
    completed_at: float | None = None
    expires_at: float | None = None
    duration: float | None = None
    size: str | None = None
    status: Literal["queued", "in_progress", "completed", "failed"] = "queued"
    progress: int | None = None
    error: str | None = None
    videos: list[VideoData] | None = None



from typing import Any, AsyncIterator, Iterator

from google.genai import Client, types
from x8._common.google_provider import GoogleProvider
from x8.core import Response
from x8.core.exceptions import BadRequestError

from .._models import (
    AllowedTools,
    FunctionCall,
    InputItem,
    OutputItem,
    OutputMessage,
    OutputMessageContent,
    OutputReasoningContentText,
    OutputText,
    Reasoning,
    ResponseCompletedEvent,
    ResponseOutputTextDeltaEvent,
    ResponseReasoningTextDeltaEvent,
    TextGenerationResult,
    TextGenerationStreamEvent,
    Tool,
    ToolChoice,
    ToolChoiceFunction,
    Usage,
)


class Google(GoogleProvider):
    vertexai: bool
    project: str | None
    location: str
    model: str
    api_key: str | None
    nparams: dict[str, Any] | None

    _client: Client
    _init: bool

    def __init__(
        self,
        vertexai: bool = True,
        project: str | None = None,
        location: str = "us-central1",
        model: str = "gemini-3-pro-preview",
        api_key: str | None = None,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        nparams: dict[str, Any] | None = None,
        **kwargs,
    ):
        self.vertexai = vertexai
        self.project = project
        self.location = location
        self.model = model
        self.api_key = api_key
        self.nparams = nparams
        super().__init__(
            service_account_info=service_account_info,
            service_account_file=service_account_file,
            access_token=access_token,
            **kwargs,
        )

    def __setup__(self, context=None):
        if self._init:
            return
        if self.api_key:
            credentials = None
        else:
            credentials = self._get_credentials()
        self._client = Client(
            vertexai=self.vertexai,
            api_key=self.api_key,
            credentials=credentials,
            project=self._get_project_or_default(self.project),
            location=self.location,
            **self.nparams or {},
        )
        self._init = True

    async def __asetup__(self, context=None):
        return self.__setup__(context)

    def generate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | Iterator[Response[TextGenerationStreamEvent]]
    ):
        args = self._convert_generate_args(
            input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        if not stream:
            response = self._client.models.generate_content(**args)
            result = self._convert_result(response)
            return Response(result=result)
        else:

            def _stream_iter() -> (
                Iterator[Response[TextGenerationStreamEvent]]
            ):
                response = self._client.models.generate_content_stream(**args)
                for event in response:
                    converted_event = self._convert_stream_event(event)
                    if converted_event is None:
                        continue
                    yield Response(result=converted_event)

            return _stream_iter()

    async def agenerate(
        self,
        input: str | list[dict[str, Any] | InputItem],
        *,
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> (
        Response[TextGenerationResult]
        | AsyncIterator[Response[TextGenerationStreamEvent]]
    ):
        args = self._convert_generate_args(
            input,
            model=model,
            instructions=instructions,
            temperature=temperature,
            top_p=top_p,
            max_output_tokens=max_output_tokens,
            tools=tools,
            tool_choice=tool_choice,
            parallel_tool_calls=parallel_tool_calls,
            max_tool_calls=max_tool_calls,
            reasoning=reasoning,
            stream=stream,
            nconfig=nconfig,
            **kwargs,
        )
        if not stream:
            response = await self._client.aio.models.generate_content(**args)
            result = self._convert_result(response)
            return Response(result=result)
        else:

            async def _poll_aiter() -> (
                AsyncIterator[Response[TextGenerationStreamEvent]]
            ):
                response = (
                    await self._client.aio.models.generate_content_stream(
                        **args
                    )
                )
                async for event in response:
                    converted_event = self._convert_stream_event(event)
                    if converted_event is None:
                        continue
                    yield Response(result=converted_event)

            return _poll_aiter()

    def _convert_stream_event(
        self, event: Any
    ) -> TextGenerationStreamEvent | None:
        # Google streaming returns GenerateContentResponse chunks, not typed
        # events. We'll synthesize internal events based on available data.
        try:
            # Prefer model_dump to get a dict we can inspect.
            e: dict[str, Any] = (
                event.model_dump() if hasattr(event, "model_dump") else event
            )
        except Exception:
            e = {}

        # Emit text delta if present.
        # Note: event.text is the concatenation of text parts in this chunk;
        # without incremental indices, we provide it as a delta unit.
        try:
            text = getattr(event, "text", None)
        except Exception:
            text = None

        if isinstance(text, str) and text:
            return ResponseOutputTextDeltaEvent(delta=text)

        # Emit reasoning delta parts if present in this chunk.
        # Parts with thought=True are considered reasoning.
        parts = e.get("parts")
        if parts and isinstance(parts, list):
            for p in parts:
                if isinstance(p, dict):
                    if p.get("thought") and isinstance(p.get("text"), str):
                        return ResponseReasoningTextDeltaEvent(
                            delta=p.get("text") or ""
                        )

        # If finish_reason is available in the first candidate, emit completed.
        candidates = e.get("candidates")
        if candidates and isinstance(candidates, list):
            c0 = candidates[0] if candidates else None
            if isinstance(c0, dict) and c0.get("finish_reason"):
                return ResponseCompletedEvent(
                    response=self._convert_result(event)
                )

        # Otherwise no event to emit for this chunk.
        return None

    def _convert_result(self, response: Any) -> TextGenerationResult:
        # Convert Google GenerateContentResponse to internal
        # TextGenerationResult.
        r: dict[str, Any] = (
            response.model_dump()
            if hasattr(response, "model_dump")
            else response
        )

        # Build output items: one assistant message (text + reasoning content),
        # plus any function calls.
        output_items: list[OutputItem] = []

        # Extract first candidate content parts.
        candidates = r.get("candidates")
        candidate0: dict[str, Any] | None = None
        if isinstance(candidates, list) and candidates:
            c0 = candidates[0]
            if isinstance(c0, dict):
                candidate0 = c0

        message_content: list[OutputMessageContent] = []
        if candidate0:
            content = candidate0.get("content")
            if isinstance(content, dict):
                parts = content.get("parts")
                if isinstance(parts, list):
                    for part in parts:
                        if not isinstance(part, dict):
                            continue
                        # Map text parts. "thought" indicates reasoning text.
                        txt = part.get("text")
                        if isinstance(txt, str):
                            if bool(part.get("thought")):
                                message_content.append(
                                    OutputReasoningContentText(text=txt)
                                )
                            else:
                                message_content.append(OutputText(text=txt))
                        # Function call parts become separate output items.
                        fc = part.get("function_call")
                        if isinstance(fc, dict) and isinstance(
                            fc.get("name"), str
                        ):
                            output_items.append(
                                FunctionCall(
                                    name=fc.get("name") or "",
                                    arguments=fc.get("args"),
                                )
                            )

        # Add assistant message if any content collected.
        if message_content:
            output_items.insert(
                0,
                OutputMessage(
                    role="assistant",
                    content=message_content,
                    status="completed",
                ),
            )

        # Usage mapping from usage_metadata.
        usage_obj: Usage | None = None
        usage = r.get("usage_metadata")
        if isinstance(usage, dict):
            input_tokens = int(usage.get("prompt_token_count") or 0)
            output_tokens = int(usage.get("candidates_token_count") or 0)
            total_tokens = int(usage.get("total_token_count") or 0)

            # Details (flatten modality counts if present)
            def _modality_details(items: Any) -> dict[str, int] | None:
                if not isinstance(items, list) or not items:
                    return None
                details: dict[str, int] = {}
                for it in items:
                    if isinstance(it, dict):
                        mod = it.get("modality")
                        cnt = it.get("token_count")
                        if mod is not None and isinstance(cnt, int):
                            details[str(mod)] = cnt
                return details or None

            usage_obj = Usage(
                input_tokens=input_tokens,
                input_tokens_details=_modality_details(
                    usage.get("prompt_tokens_details")
                ),
                output_tokens=output_tokens,
                output_tokens_details=_modality_details(
                    usage.get("candidates_tokens_details")
                ),
                total_tokens=total_tokens,
            )

        # created_at: convert datetime to epoch seconds if present.
        created_at: int | None = None
        create_time = r.get("create_time")
        try:
            if create_time is not None and hasattr(create_time, "timestamp"):
                created_at = int(create_time.timestamp())
        except Exception:
            created_at = None

        # Model info
        model_version = r.get("model_version")

        result = TextGenerationResult(
            id=r.get("response_id"),
            model=model_version,
            created_at=created_at,
            status="completed",
            error=None,
            output=output_items or None,
            usage=usage_obj,
        )
        return result

    def _convert_generate_args(
        self,
        input: str | list[dict[str, Any] | InputItem],
        model: str | None = None,
        instructions: str | None = None,
        temperature: float | None = None,
        top_p: float | None = None,
        max_output_tokens: int | None = None,
        tools: list[dict | Tool] | None = None,
        tool_choice: dict | ToolChoice | None = None,
        parallel_tool_calls: bool | None = None,
        max_tool_calls: int | None = None,
        reasoning: dict | Reasoning | None = None,
        stream: bool | None = None,
        nconfig: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        args: dict[str, Any] = {}
        args["model"] = model or self.model

        contents: list[dict[str, Any]] = []

        def _convert_input_message(msg: InputItem | dict[str, Any]) -> None:
            m: dict[str, Any]
            if isinstance(msg, dict):
                m = msg
            else:
                m = msg.to_dict()
            if m.get("type") != "message":
                return
            role = m.get("role")
            raw_content = m.get("content")

            # Map system/developer to system_instruction
            if role in ("system", "developer"):
                # Build a single Content for system_instruction
                parts: list[dict[str, Any]] = []
                if isinstance(raw_content, str):
                    parts.append({"text": raw_content})
                elif isinstance(raw_content, list):
                    for c in raw_content:
                        if not isinstance(c, dict):
                            continue
                        ct = c.get("type")
                        if ct == "input_text":
                            parts.append({"text": c.get("text")})
                        elif ct == "input_image":
                            img = c.get("image")
                            detail = c.get("detail")
                            _ = detail  # unused for now
                            # ImageData mapping
                            if isinstance(img, dict):
                                media_type = img.get("media_type")
                                content = img.get("content")
                                source = img.get("source")
                                if isinstance(content, (bytes, bytearray)):
                                    parts.append(
                                        {
                                            "inline_data": {
                                                "data": bytes(content),
                                                "mime_type": media_type
                                                or "application/octet-stream",
                                            }
                                        }
                                    )
                                elif isinstance(source, str):
                                    parts.append(
                                        {
                                            "file_data": {
                                                "file_uri": source,
                                                "mime_type": media_type
                                                or "application/octet-stream",
                                            }
                                        }
                                    )
                        elif ct == "input_file":
                            f = c.get("file")
                            if isinstance(f, dict):
                                parts.append(
                                    {
                                        "file_data": {
                                            "file_uri": (
                                                f.get("file_uri")
                                                or f.get("path")
                                                or f.get("source")
                                            ),
                                            "mime_type": (
                                                f.get("mime_type")
                                                or f.get("media_type")
                                            ),
                                        }
                                    }
                                )
                return

            # Regular user/assistant messages become contents
            role_map: dict[str, str] = {
                "user": "user",
                "assistant": "model",
                "developer": "user",
                "system": "user",
            }
            g_role = role_map.get(
                role if isinstance(role, str) else "user",
                "user",
            )
            msg_parts: list[dict[str, Any]] = []
            if isinstance(raw_content, str):
                msg_parts.append({"text": raw_content})
            elif isinstance(raw_content, list):
                for c in raw_content:
                    if not isinstance(c, dict):
                        continue
                    ct = c.get("type")
                    if ct == "input_text":
                        msg_parts.append({"text": c.get("text")})
                    elif ct == "input_image":
                        img = c.get("image")
                        detail = c.get("detail")
                        _ = detail
                        if isinstance(img, dict):
                            media_type = img.get("media_type")
                            content = img.get("content")
                            source = img.get("source")
                            if isinstance(content, (bytes, bytearray)):
                                msg_parts.append(
                                    {
                                        "inline_data": {
                                            "data": bytes(content),
                                            "mime_type": (
                                                media_type
                                                or "application/octet-stream"
                                            ),
                                        }
                                    }
                                )
                            elif isinstance(source, str):
                                msg_parts.append(
                                    {
                                        "file_data": {
                                            "file_uri": source,
                                            "mime_type": (
                                                media_type
                                                or "application/octet-stream"
                                            ),
                                        }
                                    }
                                )
                    elif ct == "input_file":
                        f = c.get("file")
                        if isinstance(f, dict):
                            msg_parts.append(
                                {
                                    "file_data": {
                                        "file_uri": (
                                            f.get("file_uri")
                                            or f.get("path")
                                            or f.get("source")
                                        ),
                                        "mime_type": (
                                            f.get("mime_type")
                                            or f.get("media_type")
                                        ),
                                    }
                                }
                            )
            contents.append({"role": g_role, "parts": msg_parts})

        # Input may be a raw string or a sequence of InputItems.
        if isinstance(input, str):
            contents.append(
                {"role": "user", "parts": [types.Part.from_text(text=input)]}
            )
        else:
            for it in input:
                _convert_input_message(it)

        args["contents"] = contents
        config: dict[str, Any] = {}
        if temperature:
            config["temperature"] = temperature
        if top_p:
            config["top_p"] = top_p
        if max_output_tokens:
            config["max_output_tokens"] = max_output_tokens
        if tools:
            config["tools"] = [self._convert_tools(tools)]
        if tool_choice:
            config["tool_config"] = self._convert_tool_choice(tool_choice)
        if instructions:
            config["system_instruction"] = instructions
        if reasoning:
            config["thinking_config"] = types.ThinkingConfig(
                include_thoughts=True
            )
        if config:
            args["config"] = config
        if nconfig:
            args.update(nconfig)
        return args

    def _convert_function(self, fn: dict) -> types.FunctionDeclaration:
        decl = types.FunctionDeclaration(
            name=fn.get("name"),
            description=fn.get("description"),
            parameters_json_schema=fn.get("parameters"),
        )
        return decl

    def _convert_tools(self, tools: list[dict | Tool]) -> list[types.Tool]:
        function_declarations: list[types.FunctionDeclaration] = []
        google_search = None
        for t in tools:
            if isinstance(t, dict):
                tool_type = t.get("type", "function")
                tool = t
            else:
                tool_type = t.type
                tool = t.to_dict()
            if tool_type == "function":
                decl = self._convert_function(tool)
                function_declarations.append(decl)
            elif tool_type == "web_search":
                google_search = types.GoogleSearch()
            else:
                raise BadRequestError(f"Unsupported tool type: {tool_type}")
        return [
            types.Tool(
                function_declarations=function_declarations or None,
                google_search=google_search,
            )
        ]

    def _convert_tool_choice(
        self,
        choice: dict | ToolChoice | None,
    ) -> types.ToolConfig | None:
        if choice is None:
            return None

        if isinstance(choice, ToolChoiceFunction) or isinstance(
            choice, AllowedTools
        ):
            inp = choice.to_dict()
        else:
            inp = choice

        choice_mode_map: dict = {
            "none": types.FunctionCallingConfigMode.NONE,
            "auto": types.FunctionCallingConfigMode.AUTO,
            "required": types.FunctionCallingConfigMode.ANY,
        }
        if inp in ("none", "auto", "required"):
            return types.ToolConfig(
                function_calling_config=types.FunctionCallingConfig(
                    mode=choice_mode_map[inp]
                )
            )
        if isinstance(inp, dict):
            if inp.get("type") == "function":
                name = inp.get("name")
                if not name:
                    raise BadRequestError(
                        "Function tool_choice must specify a name"
                    )
                return types.ToolConfig(
                    function_calling_config=types.FunctionCallingConfig(
                        mode=types.FunctionCallingConfigMode.ANY,
                        allowed_function_names=[name],
                    )
                )
            elif inp.get("type") == "allowed_tools":
                mode = inp.get("mode")
                allowed = inp.get("tools")
                allowed_function_names = (
                    [t.get("name") for t in allowed]
                    if isinstance(allowed, list)
                    else None
                )
                return types.ToolConfig(
                    function_calling_config=types.FunctionCallingConfig(
                        mode=choice_mode_map[mode],
                        allowed_function_names=allowed_function_names,
                    )
                )

        raise BadRequestError(f"Unsupported tool_choice: {type(choice)}")



from typing import Any

from common.secrets import get_secrets

from x8.ai.video_generation import VideoGeneration

secrets = get_secrets()


class VideoGenerationProvider:
    OPENAI = "openai"


provider_parameters: dict[str, dict[str, Any]] = {
    VideoGenerationProvider.OPENAI: {
        "api_key": secrets["ai-openai-api-key"],
    },
}


def get_component(provider_type: str):
    parameters = provider_parameters[provider_type]
    component = VideoGeneration(
        __unpack__=True,
        __provider__=dict(
            type=provider_type,
            parameters=parameters,
        ),
    )
    return component



__all__ = ["AzureServiceBus"]

from typing import Any

from x8.messaging._common import MessagingMode
from x8.messaging._common.azure_service_bus import AzureServiceBusBase


class AzureServiceBus(AzureServiceBusBase):
    def __init__(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        fully_qualified_namespace: str | None = None,
        connection_string: str | None = None,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            topic:
                Service Bus topic name.
            subscription:
                Service Bus subscription name.
            fully_qualified_namespace:
                The fully qualified host name for the Service Bus namespace.
                The namespace format is:
                    `<yournamespace>.servicebus.windows.net`.
            connection_string:
                The connection string of a Service Bus.
            credential_type:
                Azure credential type.
            tenant_id:
                Azure tenant id for client_secret credential type.
            client_id:
                Azure client id for client_secret credential type.
            client_secret:
                Azure client secret for client_secret credential type.
            certificate_path:
                Certificate path for certificate credential type.
            nparams:
                Native parameters to Service Bus client.
        """
        super().__init__(
            mode=MessagingMode.PUBSUB,
            topic=topic,
            subscription=subscription,
            fully_qualified_namespace=fully_qualified_namespace,
            connection_string=connection_string,
            credential_type=credential_type,
            tenant_id=tenant_id,
            client_id=client_id,
            client_secret=client_secret,
            certificate_path=certificate_path,
            nparams=nparams,
            **kwargs,
        )

    def __supports__(self, feature: str) -> bool:
        return True



from typing import Any

from common.secrets import get_secrets
from storage.document_store import get_component as get_ds_component
from storage.object_store import get_component as get_os_component

from x8.storage.key_value_store import KeyValueStore

secrets = get_secrets()


class KeyValueStoreProvider:
    REDIS = "redis"
    REDIS_SIMPLE = "redis_simple"
    MEMCACHED = "memcached"
    MEMORY = "memory"
    SQLITE = "sqlite"
    POSTGRESQL = "postgresql"
    DS_AMAZON_DYNAMODB = "ds_amazon_dynamodb"
    DS_AZURE_COSMOS_DB = "ds_azure_cosmos_db"
    DS_GOOGLE_FIRESTORE = "ds_google_firestore"
    DS_MONGODB = "ds_mongodb"
    DS_POSTGRESQL = "ds_postgresql"
    DS_REDIS = "ds_redis"
    DS_SQLITE = "ds_sqlite"
    DS_MEMORY = "ds_memory"
    OS_AMAZON_S3 = "os_amazon_s3"
    OS_AZURE_BLOB_STORAGE = "os_azure_blob_storage"
    OS_GOOGLE_CLOUD_STORAGE = "os_google_cloud_storage"
    OS_FILE_SYSTEM = "os_file_system"


provider_parameters: dict[str, dict[str, Any]] = {
    KeyValueStoreProvider.REDIS: {
        "host": secrets["storage-document-redis-host"],
        "port": secrets["storage-document-redis-port"],
        "password": secrets["storage-document-redis-password"],
    },
    KeyValueStoreProvider.REDIS_SIMPLE: {
        "host": secrets["storage-document-redis-host"],
        "port": secrets["storage-document-redis-port"],
        "password": secrets["storage-document-redis-password"],
    },
    KeyValueStoreProvider.MEMCACHED: {
        "host": "127.0.0.1",
        "port": 11211,
    },
    KeyValueStoreProvider.MEMORY: {},
    KeyValueStoreProvider.SQLITE: {},
    KeyValueStoreProvider.POSTGRESQL: {
        "connection_string": secrets[
            "storage-document-postgresql-connection-string"
        ]
    },
}


def get_component(provider_type: str, type="binary", collection: str = "test"):
    base_provider_type = provider_type
    if provider_type.startswith("ds"):
        splits = provider_type.split("_")
        base_provider_type = "document_store_provider"
        parameters = {
            "store": get_ds_component("_".join(splits[1:]), "kvtest")
        }
    elif provider_type.startswith("os"):
        splits = provider_type.split("_")
        base_provider_type = "object_store_provider"
        parameters = {"store": get_os_component("_".join(splits[1:]))}
    else:
        parameters = provider_parameters[provider_type]

    component = KeyValueStore(
        type=type,
        collection=collection,
        __provider__=dict(
            type=base_provider_type,
            parameters=parameters,
        ),
    )
    return component



from typing import Any

from common.secrets import get_secrets

from x8.storage.search_store import SearchStore

secrets = get_secrets()


class SearchStoreProvider:
    ELASTICSEARCH = "elasticsearch"


provider_parameters: dict[str, dict[str, Any]] = {
    SearchStoreProvider.ELASTICSEARCH: {
        "hosts": "http://localhost:9200",
    },
}


def get_component(provider_type: str, collection: str = "test"):
    component = SearchStore(
        collection=collection,
        __provider__=dict(
            type=provider_type,
            parameters=provider_parameters[provider_type],
        ),
    )
    return component



from x8.compute._common._image_helper import aget_images, get_images
from x8.compute._common._models import ImageMap
from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import Containerizer
from x8.core import RunContext
from x8.core.exceptions import BadRequestError

from ._models import Container, EnvVar, ServiceDefinition, ServiceOverlay


def map_images(
    containers: list[Container],
    images: list[ImageMap],
    containerizer: Containerizer | None = None,
    container_registry: ContainerRegistry | None = None,
    run_context: RunContext = RunContext(),
) -> list[str]:
    image_uris = get_images(
        images=images,
        containerizer=containerizer,
        container_registry=container_registry,
        run_context=run_context,
    )
    container_images = []
    for container in containers:
        found = False
        for i, image in enumerate(images):
            if image.name == container.name:
                container_images.append(image_uris[i])
                found = True
                break
        if not found:
            if container.image:
                container_images.append(container.image)
            else:
                raise BadRequestError(
                    (
                        f"Container image not found for container: "
                        f"{container.name}"
                    )
                )
    return container_images


async def amap_images(
    containers: list[Container],
    images: list[ImageMap],
    containerizer: Containerizer | None = None,
    container_registry: ContainerRegistry | None = None,
    run_context: RunContext = RunContext(),
) -> list[str]:
    image_uris = await aget_images(
        images=images,
        containerizer=containerizer,
        container_registry=container_registry,
        run_context=run_context,
    )
    container_images = []
    for container in containers:
        found = False
        for i, image in enumerate(images):
            if image.name == container.name:
                container_images.append(image_uris[i])
                found = True
                break
        if not found:
            if container.image:
                container_images.append(container.image)
            else:
                raise BadRequestError(
                    (
                        f"Container image not found for container: "
                        f"{container.name}"
                    )
                )
    return container_images


def get_env(base_env: list[EnvVar], overlay: list[EnvVar]) -> list[EnvVar]:
    env_dict = {item.name: item.value for item in base_env}
    for item in overlay:
        env_dict[item.name] = item.value
    return [EnvVar(name=k, value=v) for k, v in env_dict.items()]


def merge_service_overlay(
    service: ServiceDefinition,
    overlay: ServiceOverlay | None,
) -> ServiceDefinition:
    if overlay is None:
        return service
    if overlay.containers:
        for container_override in overlay.containers:
            for container in service.containers:
                if container.name == container_override.name:
                    container.env = get_env(
                        base_env=container.env,
                        overlay=container_override.env,
                    )
    return service


def requires_container_registry(
    service: ServiceDefinition,
) -> bool:
    for container in service.containers:
        if not container.image:
            return True
    return False



from x8.compute.container_registry import ContainerRegistry
from x8.compute.containerizer import BuildConfig, Containerizer
from x8.core import RunContext
from x8.core.exceptions import BadRequestError

from ._models import ImageMap


def get_image(
    image_map: ImageMap,
    containerizer: Containerizer | None = None,
    container_registry: ContainerRegistry | None = None,
    run_context: RunContext = RunContext(),
) -> str:
    source = None
    if image_map.handle or image_map.component:
        handle = image_map.handle
        if not handle and image_map.component:
            handle = image_map.component.__handle__
        if not handle:
            raise BadRequestError(
                "Container handle or component is required to get image."
            )
        if containerizer is None:
            raise BadRequestError(
                "Containerizer is required to get image from component."
            )
        prepare_res = containerizer.prepare(
            handle=handle,
            config=image_map.prepare,
            run_context=run_context,
        )
        source = prepare_res.result.source
    elif image_map.source:
        source = image_map.source
    local_image = None
    if source:
        if containerizer is None:
            raise BadRequestError(
                "Containerizer is required to get image from source."
            )
        build_config = image_map.build
        if not build_config:
            build_config = BuildConfig(image_name=image_map.name)
        if not build_config.image_name:
            build_config.image_name = image_map.name
        _ = containerizer.build(source=source, config=build_config)
        local_image = build_config.image_name
    elif image_map.local_image:
        local_image = image_map.local_image

    if local_image:
        if container_registry is None:
            raise BadRequestError(
                "Container registry is required to push local image."
            )
        push_res = container_registry.push(image_name=local_image)
        image_uri = push_res.result.image_uri
    else:
        raise BadRequestError(
            "Container image or local image or source is required."
        )
    return image_uri


async def aget_image(
    image_map: ImageMap,
    containerizer: Containerizer | None = None,
    container_registry: ContainerRegistry | None = None,
    run_context: RunContext = RunContext(),
) -> str:
    source = None
    if image_map.handle or image_map.component:
        handle = image_map.handle
        if not handle and image_map.component:
            handle = image_map.component.__handle__
        if not handle:
            raise BadRequestError(
                "Container handle or component is required to get image."
            )
        if containerizer is None:
            raise BadRequestError(
                "Containerizer is required to get image from component."
            )
        prepare_res = await containerizer.aprepare(
            handle=handle,
            config=image_map.prepare,
            run_context=run_context,
        )
        source = prepare_res.result.source
    elif image_map.source:
        source = image_map.source
    local_image = None
    if source:
        if containerizer is None:
            raise BadRequestError(
                "Containerizer is required to get image from source."
            )
        build_config = image_map.build
        if not build_config:
            build_config = BuildConfig(image_name=image_map.name)
        if not build_config.image_name:
            build_config.image_name = image_map.name
        _ = await containerizer.abuild(source=source, config=build_config)
        local_image = build_config.image_name
    elif image_map.local_image:
        local_image = image_map.local_image

    if local_image:
        if container_registry is None:
            raise BadRequestError(
                "Container registry is required to push image."
            )
        push_res = await container_registry.apush(image_name=local_image)
        image_uri = push_res.result.image_uri
    else:
        raise BadRequestError(
            "Container image or local image or source is required."
        )
    return image_uri


def get_images(
    images: list[ImageMap],
    containerizer: Containerizer | None = None,
    container_registry: ContainerRegistry | None = None,
    run_context: RunContext = RunContext(),
) -> list[str]:
    result: list[str] = []
    for image_map in images:
        res = get_image(
            image_map=image_map,
            containerizer=containerizer,
            container_registry=container_registry,
            run_context=run_context,
        )
        result.append(res)
    return result


async def aget_images(
    images: list[ImageMap],
    containerizer: Containerizer | None = None,
    container_registry: ContainerRegistry | None = None,
    run_context: RunContext = RunContext(),
) -> list[str]:
    result: list[str] = []
    for image_map in images:
        res = await aget_image(
            image_map=image_map,
            containerizer=containerizer,
            container_registry=container_registry,
            run_context=run_context,
        )
        result.append(res)
    return result



# type: ignore
import asyncio
import os
import time
from datetime import datetime, timezone

import pytest

from x8.storage.config_store import DEFAULT_LABEL, ConfigItem, NotFoundError

from ._data import configs
from ._providers import ConfigStoreProvider
from ._sync_and_async_client import ConfigStoreSyncAndAsyncClient

if os.name == "nt":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

queries = [
    {
        "query": None,
        "result": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    },
    {
        "query": "starts_with($id, '')",
        "result": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
    },
    {
        "query": "starts_with($id, '/root/')",
        "result": [0, 1, 2, 3, 6, 7, 9, 10],
    },
    {
        "query": "starts_with($id, '/root/test1')",
        "result": [0, 1, 6, 9],
    },
    {
        "query": f"$label = '{DEFAULT_LABEL}'",
        "result": [0, 1, 2, 3, 4, 5],
    },
    {
        "query": "$label = 'label1'",
        "result": [6, 7, 8],
    },
    {
        "query": "$label = 'label2'",
        "result": [9, 10, 11],
    },
    {
        "query": f"starts_with($id, '') and $label = '{DEFAULT_LABEL}'",
        "result": [0, 1, 2, 3, 4, 5],
    },
    {
        "query": "starts_with($id, '') and $label = 'label1'",
        "result": [6, 7, 8],
    },
    {
        "query": f"starts_with($id, '/root/') and $label = '{DEFAULT_LABEL}'",
        "result": [0, 1, 2, 3],
    },
    {
        "query": "starts_with($id, '/root/') and $label = 'label2'",
        "result": [9, 10],
    },
    {
        "query": f"""starts_with($id, '/root/test1')
                and $label = '{DEFAULT_LABEL}'""",
        "result": [0, 1],
    },
    {
        "query": "starts_with($id, '/root/test1') and $label = 'label1'",
        "result": [6],
    },
    {
        "query": "starts_with($id, '/root/test2') and $label = 'label2'",
        "result": [10],
    },
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        ConfigStoreProvider.AWS_PARAMETER_STORE,
        ConfigStoreProvider.AZURE_APP_CONFIGURATION,
        ConfigStoreProvider.GOOGLE_RUNTIME_CONFIGURATOR,
        ConfigStoreProvider.DS_AMAZON_DYNAMODB,
        ConfigStoreProvider.DS_AZURE_COSMOS_DB,
        ConfigStoreProvider.DS_GOOGLE_FIRESTORE,
        ConfigStoreProvider.DS_MONGODB,
        ConfigStoreProvider.DS_POSTGRESQL,
        ConfigStoreProvider.DS_REDIS,
        ConfigStoreProvider.DS_MEMORY,
        ConfigStoreProvider.DS_SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_put_get_delete(provider_type: str, async_call: bool):
    config = configs[0]
    key = get_key(config)
    value = config["value"]
    client = ConfigStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )
    await cleanup_config(provider_type, key, client=client)

    with pytest.raises(NotFoundError):
        await client.get(key=key)

    response = await client.put(key=key, value=value)
    result = response.result
    assert_config_item(result, config)

    await wait_for_put(provider_type)

    response = await client.get(key=key)
    result = response.result
    assert_config_item(result, config)

    config["value"] = "new value"
    response = await client.put(key=key, value=config["value"])
    result = response.result
    assert_config_item(result, config)

    await wait_for_put(provider_type)

    response = await client.get(key=key)
    result = response.result
    assert_config_item(result, config)

    response = await client.delete(key=key)
    result = response.result
    assert result is None

    await wait_for_delete(provider_type)

    with pytest.raises(NotFoundError):
        await client.get(key=key)

    with pytest.raises(NotFoundError):
        await client.delete(key=key)
    await client.close()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        ConfigStoreProvider.AWS_PARAMETER_STORE,
        ConfigStoreProvider.AZURE_APP_CONFIGURATION,
        ConfigStoreProvider.GOOGLE_RUNTIME_CONFIGURATOR,
        ConfigStoreProvider.DS_AMAZON_DYNAMODB,
        ConfigStoreProvider.DS_AZURE_COSMOS_DB,
        ConfigStoreProvider.DS_GOOGLE_FIRESTORE,
        ConfigStoreProvider.DS_MONGODB,
        ConfigStoreProvider.DS_POSTGRESQL,
        ConfigStoreProvider.DS_REDIS,
        ConfigStoreProvider.DS_MEMORY,
        ConfigStoreProvider.DS_SQLITE,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_query_count(provider_type: str, async_call: bool):
    client = ConfigStoreSyncAndAsyncClient(
        provider_type=provider_type, async_call=async_call
    )

    for config in configs:
        key = get_key(config)
        value = config["value"]
        response = await client.put(key=key, value=value)
        result = response.result
        assert_config_item(result, config)

    await wait_for_put(provider_type)

    for query in queries:
        response = await client.query(query["query"])
        result = response.result
        assert_config_list(result.items, query["result"])
        response = await client.count(query["query"])
        result = response.result
        assert result == len(query["result"])

    for config in configs:
        key = get_key(config)
        await client.delete(key=key)

    await wait_for_delete(provider_type)

    await client.close()


def get_key(config):
    if config["label"] is None:
        return config["id"]
    return {"id": config["id"], "label": config["label"]}


def assert_config_list(items: list[ConfigItem], sorted_index: list[int]):
    assert len(items) == len(sorted_index)
    for i in range(len(sorted_index)):
        assert_config_item(items[i], configs[sorted_index[i]])


def assert_config_item(item: ConfigItem, config):
    assert item.key.id == config["id"]
    if config["label"] is None:
        assert item.key.label == DEFAULT_LABEL
    else:
        assert item.key.label == config["label"]
    assert item.value == config["value"]
    if (
        item.properties is not None
        and item.properties.updated_time is not None
    ):
        diff_time_seconds = (
            datetime.now(timezone.utc).timestamp()
            - item.properties.updated_time
        )
        assert diff_time_seconds < 1800


async def cleanup_config(
    provider_type, key, client: ConfigStoreSyncAndAsyncClient
):
    try:
        await client.delete(key=key)
    except NotFoundError:
        return

    await wait_for_delete(provider_type)


async def wait_for_delete(provider_type):
    if provider_type == ConfigStoreProvider.AWS_PARAMETER_STORE:
        time.sleep(1)


async def wait_for_put(provider_type):
    if provider_type == ConfigStoreProvider.AWS_PARAMETER_STORE:
        time.sleep(5)



# type: ignore
import os

import pytest

from x8.content.live_stream import LiveStreamParam

from ._providers import LiveStreamProvider
from ._sync_and_async_client import LiveStreamSyncAndAsyncClient


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "provider_type",
    [
        LiveStreamProvider.AV,
    ],
)
@pytest.mark.parametrize(
    "async_call",
    [False, True],
)
async def test_single(provider_type: str, async_call: bool):
    live_stream = LiveStreamSyncAndAsyncClient(
        provider_type=provider_type,
        async_call=async_call,
        live_stream_param=LiveStreamParam(
            url="https://newsmax.codeautoplaycdn2.workers.dev/"
        ),
    )

    info = await live_stream.get_info()
    print(info)
    raise


def get_file_path(file):
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "data",
        file,
    )



import inspect
import re
from collections.abc import AsyncIterator, Iterator
from types import UnionType
from typing import Any, Callable, Union, get_args, get_origin, get_type_hints

from ._models import (
    APIAuth,
    ArgMapping,
    ArgSource,
    ArgSourceType,
    ComponentMapping,
    OperationMapping,
    ResponseMapping,
)


class ArgInfo:
    name: str
    type: Any
    default: Any | None
    required: bool
    source: ArgSource

    def __init__(
        self,
        name: str,
        type: Any,
        default: Any | None,
        required: bool,
        source: ArgSource,
    ):
        self.name = name
        self.type = type
        self.default = default
        self.required = required
        self.source = source


class OperationInfo:
    name: str
    path: str
    method: Callable[..., Any]
    http_method: str
    status_code: int
    return_type: Any | None
    is_return_iterator: bool
    path_params: list[str]
    auth: APIAuth | None
    args: list[ArgInfo]
    response: ResponseMapping | None
    operation_mapping: OperationMapping | None

    def __init__(self, name: str, method: Callable[..., Any]):
        self.name = name
        self.path = f"/{self.name}"
        self.method = method
        self.http_method = "POST"
        self.status_code = 200
        self.path_params = []
        self.auth = None
        self.response = None
        self.operation_mapping = None


def get_components(
    component_mappings: list[ComponentMapping],
) -> list[ComponentMapping]:
    """
    Get components from component mappings
    """
    components: list[ComponentMapping] = []
    for component_mapping in component_mappings:
        components.append(get_component(component_mapping))
    return components


def get_component(
    component_mapping: ComponentMapping,
) -> ComponentMapping:
    config = getattr(component_mapping.component, "__config__", {})
    api_config = config.get("api", {})
    component_mapping.prefix = component_mapping.prefix or api_config.get(
        "prefix", None
    )
    component_mapping.tags = component_mapping.tags or api_config.get(
        "tags", []
    )
    component_mapping.args = component_mapping.args or api_config.get(
        "args", []
    )
    return component_mapping


def get_operations(
    component_mapping: ComponentMapping,
    auth: APIAuth | None = None,
) -> dict[str, OperationInfo]:
    """
    Get operations from component
    """
    operations: dict[str, OperationInfo] = {}
    # Get all the sync methods first.
    for method_name, method in inspect.getmembers(
        component_mapping.component, predicate=inspect.ismethod
    ):
        if (
            callable(method)
            and not inspect.iscoroutinefunction(method)
            and hasattr(method, "__operation__")
        ):
            operations[method_name] = OperationInfo(method_name, method)

    # Next, iterate over async methods.
    # If there is an async equivalent of a sync method,
    # then resolve based on the resolution policy.
    for method_name, method in inspect.getmembers(
        component_mapping.component, predicate=inspect.iscoroutinefunction
    ):
        if callable(method) and hasattr(method, "__operation__"):
            if method_name.startswith("a") and method_name[1:] in operations:
                if component_mapping.sync_async_resolution == "async":
                    operations[method_name[1:]] = OperationInfo(
                        method_name[1:], method
                    )
                elif component_mapping.sync_async_resolution == "both":
                    operations[method_name] = OperationInfo(
                        method_name, method
                    )
            else:
                if (
                    method_name.startswith("a")
                    and component_mapping.remove_async_prefix
                ):
                    operations[method_name[1:]] = OperationInfo(
                        method_name[1:], method
                    )
                else:
                    operations[method_name] = OperationInfo(
                        method_name, method
                    )
    # Remove operations that are suppressed.
    for operation in component_mapping.operations:
        if operation.supress:
            operations.pop(operation.name, None)

    # Update operation info with mappings.
    for op_name, op_info in operations.items():
        op_info.auth = component_mapping.auth or auth
        op_info.return_type = get_return_type(op_info.method)
        op_info.is_return_iterator = get_is_iterator_type(op_info.return_type)
        if component_mapping.predict_http_method:
            op_info.http_method = predict_http_method(op_name)
        else:
            op_info.http_method = "POST"
        op_mapping = get_operation_mapping(component_mapping, op_name)
        if not op_mapping and hasattr(op_info.method, "__config__"):
            config = getattr(op_info.method, "__config__", {})
            if "api" in config:
                op_mapping = OperationMapping(
                    name=op_name,
                    **config["api"],
                )
        arg_mappings = []
        if op_mapping:
            op_info.path = (
                op_mapping.path
                if op_mapping.path is not None
                else op_info.path
            )
            op_info.http_method = op_mapping.method or op_info.http_method
            op_info.status_code = op_mapping.status or op_info.status_code
            op_info.path_params = re.findall(r"\{(.*?)\}", op_info.path)
            op_info.path_params.extend(
                re.findall(r"\{(.*?)\}", component_mapping.prefix or "")
            )
            op_info.auth = op_mapping.auth or op_info.auth
            op_info.response = op_mapping.response or op_info.response
            op_info.operation_mapping = op_mapping
            arg_mappings = op_mapping.args

        op_info.http_method = op_info.http_method.upper()
        for arg_mapping in component_mapping.args:
            if arg_mapping.name not in [a.name for a in arg_mappings]:
                arg_mappings.append(arg_mapping)
        op_info.args = get_arg_info(
            op_info,
            arg_mappings,
        )
    return operations


def get_operation_mapping(
    component_mapping: ComponentMapping,
    operation_name: str,
) -> OperationMapping | None:
    for operation in component_mapping.operations:
        if operation.name == operation_name:
            return operation
    return None


def get_arg_mapping(
    arg_mappings: list[ArgMapping],
    arg_name: str,
) -> ArgMapping | None:
    for arg_mapping in arg_mappings:
        if arg_mapping.name == arg_name:
            return arg_mapping
    return None


def get_return_type(method: Any) -> Any:
    sig = inspect.signature(method)
    t = (
        sig.return_annotation
        if sig.return_annotation != inspect.Signature.empty
        else None
    )
    origin = get_origin(t)
    if origin in (Union, UnionType):
        args = get_args(t)
        for a in args:
            o = get_origin(a)
            if o in (Iterator, AsyncIterator):
                return get_args(a)[0]
            else:
                return a
    return t


def get_is_iterator_type(t: Any) -> bool:
    origin = get_origin(t)
    return origin in (Iterator, AsyncIterator)


def predict_http_method(operation_name: str) -> str:
    mapping = {
        "GET": (
            "get",
            "aget",
            "list",
            "alist",
            "query",
            "aquery",
            "fetch",
            "afetch",
            "retrieve",
            "aretrieve",
            "find",
            "afind",
            "search",
            "asearch",
            "read",
            "aread",
        ),
        "PUT": (
            "put",
            "aput",
            "update",
            "aupdate",
            "replace",
            "areplace",
            "modify",
            "amodify",
            "save",
            "asave",
        ),
        "DELETE": (
            "delete",
            "adelete",
            "remove",
            "aremove",
            "destroy",
            "adestroy",
            "purge",
            "apurge",
            "clear",
            "aclear",
        ),
        "PATCH": (
            "patch",
            "apatch",
        ),
    }
    for method, prefixes in mapping.items():
        if operation_name.startswith(prefixes):
            return method
    return "POST"


def get_arg_info(
    op_info: OperationInfo,
    arg_mappings: list[ArgMapping],
) -> list[ArgInfo]:
    sig = inspect.signature(op_info.method)
    type_hints = get_type_hints(op_info.method)
    args: list[ArgInfo] = []
    for name, param in sig.parameters.items():
        if name == "self":
            continue
        if name == "kwargs" and param.kind == inspect.Parameter.VAR_KEYWORD:
            continue
        arg_mapping = get_arg_mapping(arg_mappings, name)
        if arg_mapping and arg_mapping.source:
            source = arg_mapping.source
        else:
            if name in op_info.path_params:
                source = ArgSource(type=ArgSourceType.PATH, field=name)
            elif op_info.http_method in ["POST", "PUT", "PATCH"]:
                source = ArgSource(type=ArgSourceType.BODY, field=name)
            else:
                source = ArgSource(type=ArgSourceType.QUERY, field=name)
        type = type_hints.get(name, Any)
        arg_info = ArgInfo(
            name=name,
            type=type,
            default=(
                param.default
                if param.default != inspect.Parameter.empty
                else None
            ),
            required=param.default == inspect.Parameter.empty,
            source=source,
        )
        args.append(arg_info)
    return args



from ._models import (
    DEFAULT_SUBSCRIPTION_NAME,
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingMode,
    QueueConfig,
    QueueInfo,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)
from ._operation import MessagingOperation
from ._operation_parser import MessagingOperationParser

__all__ = [
    "MessageBatch",
    "MessageKey",
    "MessageItem",
    "MessageProperties",
    "MessagePullConfig",
    "MessagePutConfig",
    "MessageValueType",
    "MessagingOperation",
    "MessagingOperationParser",
    "QueueConfig",
    "SubscriptionConfig",
    "TopicConfig",
    "QueueInfo",
    "TopicInfo",
    "SubscriptionInfo",
    "MessagingMode",
    "DEFAULT_SUBSCRIPTION_NAME",
]



from x8.core.exceptions import ConflictError, NotFoundError
from x8.messaging._common import (
    MessageBatch,
    MessageItem,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    QueueConfig,
    QueueInfo,
)

from ._feature import QueueFeature
from .component import Queue

__all__ = [
    "MessageBatch",
    "MessageItem",
    "MessageProperties",
    "MessagePullConfig",
    "MessagePutConfig",
    "MessageValueType",
    "Queue",
    "QueueConfig",
    "QueueInfo",
    "QueueFeature",
    "NotFoundError",
    "ConflictError",
]



import base64
import os
import subprocess
from typing import Any

import boto3
from botocore.exceptions import ClientError
from x8.compute.containerizer import DEFAULT_PLATFORM
from x8.core import Context, Provider, Response
from x8.core.exceptions import BadRequestError

from .._models import ContainerRegistryItem, ContainerRegistryItemDigest


class AmazonElasticContainerRegistry(Provider):
    region: str
    registry_id: str | None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    profile_name: str | None
    platform: str
    nparams: dict[str, Any]

    _client: Any
    _init: bool = False
    _shell: bool = False
    _account_id: str | None = None

    def __init__(
        self,
        region: str = "us-west-2",
        registry_id: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        profile_name: str | None = None,
        platform: str = DEFAULT_PLATFORM,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            region:
                AWS region where the ECR registry is located.
            registry_id:
                AWS account ID that owns the ECR registry.
                If None, uses the default account.
            aws_access_key_id:
                AWS access key ID for authentication.
            aws_secret_access_key:
                AWS secret access key for authentication.
            aws_session_token:
                AWS session token for temporary credentials.
            profile_name:
                AWS profile name to use for authentication.
            platform:
                Platform to use for the container when pulling.
            nparams:
                Additional parameters for the ECR client.
        """
        self.region = region
        self.registry_id = registry_id
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.profile_name = profile_name
        self.platform = platform
        self.nparams = nparams
        self._init = False
        self._ainit = False
        self._client = None
        self._aclient = None
        self._shell = os.name == "nt"
        super().__init__(**kwargs)

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        session_kwargs = {}
        if self.aws_access_key_id:
            session_kwargs["aws_access_key_id"] = self.aws_access_key_id
        if self.aws_secret_access_key:
            session_kwargs["aws_secret_access_key"] = (
                self.aws_secret_access_key
            )
        if self.aws_session_token:
            session_kwargs["aws_session_token"] = self.aws_session_token
        if self.profile_name:
            session_kwargs["profile_name"] = self.profile_name

        session = boto3.Session(**session_kwargs)
        self._client = session.client(
            "ecr", region_name=self.region, **self.nparams
        )

        if self.registry_id:
            self._account_id = self.registry_id
        else:
            sts = session.client("sts")
            self._account_id = sts.get_caller_identity()["Account"]
        self._init = True

    def _get_registry_url(self) -> str:
        return f"{self._account_id}.dkr.ecr.{self.region}.amazonaws.com"

    def _get_full_image_path(
        self, image_name: str, tag: str | None = None
    ) -> str:
        image = f"{self._get_registry_url()}/{image_name}"
        if tag:
            image += f":{tag}"
        return image

    def _docker_login(self) -> None:
        """Authenticate Docker with ECR."""

        # Get authorization token
        response = self._client.get_authorization_token(
            registryIds=[self._account_id]
        )

        auth_data = response["authorizationData"][0]
        token = auth_data["authorizationToken"]
        endpoint = auth_data["proxyEndpoint"]

        # Decode the token
        username, password = base64.b64decode(token).decode().split(":", 1)

        # Login to Docker
        subprocess.run(
            [
                "docker",
                "login",
                "--username",
                username,
                "--password-stdin",
                endpoint,
            ],
            input=password,
            text=True,
            shell=self._shell,
            check=True,
        )

    def push(self, image_name: str) -> Response[ContainerRegistryItem]:
        self.__setup__()
        self._docker_login()

        full_path = self._get_full_image_path(image_name)
        subprocess.run(
            ["docker", "tag", image_name, full_path],
            shell=self._shell,
            check=True,
        )
        repo_name = image_name.split(":")[0]
        try:
            self._client.describe_repositories(repositoryNames=[repo_name])
        except ClientError as e:
            if e.response["Error"]["Code"] == "RepositoryNotFoundException":
                self._client.create_repository(repositoryName=repo_name)
            else:
                raise
        subprocess.run(
            ["docker", "push", full_path], shell=self._shell, check=True
        )

        result = ContainerRegistryItem(
            image_name=image_name, image_uri=full_path
        )
        return Response(result=result)

    def pull(
        self, image_name: str, tag: str | None = None
    ) -> Response[ContainerRegistryItem]:
        self._docker_login()

        full_path = self._get_full_image_path(image_name, tag)
        subprocess.run(
            ["docker", "pull", "--platform", self.platform, full_path],
            shell=self._shell,
            check=True,
        )

        result = ContainerRegistryItem(
            image_name=image_name, image_uri=full_path
        )
        return Response(result=result)

    def tag(
        self,
        image_name: str,
        tag: str,
        digest: str | None = None,
    ) -> Response[ContainerRegistryItemDigest]:
        self.__setup__()
        self._docker_login()

        if digest is None:
            try:
                images = self._client.describe_images(
                    repositoryName=image_name,
                    imageIds=[{"imageTag": "latest"}],
                )["imageDetails"]
            except Exception:
                images = self._client.describe_images(
                    repositoryName=image_name, maxResults=1
                )["imageDetails"]
            if not images:
                raise BadRequestError(
                    f"No images found for {image_name} with tag 'latest'."
                )
            digest = images[0]["imageDigest"]

        source_path = f"{self._get_full_image_path(image_name)}@{digest}"
        tagged_path = f"{self._get_full_image_path(image_name)}:{tag}"

        subprocess.run(
            ["docker", "pull", "--platform", self.platform, source_path],
            shell=self._shell,
            check=True,
        )
        subprocess.run(
            ["docker", "tag", source_path, tagged_path],
            shell=self._shell,
            check=True,
        )
        subprocess.run(
            ["docker", "push", tagged_path], shell=self._shell, check=True
        )

        result = ContainerRegistryItemDigest(
            image_uri=tagged_path,
            digest=digest,
            upload_time=None,
            image_size_bytes=None,
            tags=[tag],
        )
        return Response(result=result)

    def delete(
        self,
        image_name: str,
        digest: str | None = None,
        tag: str | None = None,
    ) -> Response[None]:
        self.__setup__()

        if digest is None and tag is None:
            # Delete entire repository
            self._client.delete_repository(
                repositoryName=image_name, force=True
            )
        else:
            # Delete specific image
            image_ids = []
            if digest:
                image_ids.append({"imageDigest": digest})
            if tag:
                image_ids.append({"imageTag": tag})

            self._client.batch_delete_image(
                repositoryName=image_name, imageIds=image_ids
            )
        return Response(result=None)

    def list_images(self) -> Response[list[ContainerRegistryItem]]:
        self.__setup__()

        paginator = self._client.get_paginator("describe_repositories")
        items = []

        for page in paginator.paginate():
            for repo in page["repositories"]:
                repo_name = repo["repositoryName"]
                items.append(
                    ContainerRegistryItem(
                        image_name=repo_name,
                        image_uri=self._get_full_image_path(repo_name),
                    )
                )

        result = items
        return Response(result=result)

    def get_digests(
        self, image_name: str
    ) -> Response[list[ContainerRegistryItemDigest]]:
        self.__setup__()

        paginator = self._client.get_paginator("describe_images")
        digests = []

        for page in paginator.paginate(repositoryName=image_name):
            for image_detail in page["imageDetails"]:
                upload_time = None
                if "imagePushedAt" in image_detail:
                    upload_time = image_detail["imagePushedAt"].timestamp()

                image_uri = (
                    f"{self._get_full_image_path(image_name)}"
                    f"@{image_detail['imageDigest']}"
                )

                digests.append(
                    ContainerRegistryItemDigest(
                        image_uri=image_uri,
                        digest=image_detail["imageDigest"],
                        upload_time=upload_time,
                        image_size_bytes=image_detail.get("imageSizeInBytes"),
                        tags=image_detail.get("imageTags"),
                    )
                )

        result = digests
        return Response(result=result)

    def close(self) -> Response[None]:
        self._init = False
        return Response(result=None)



class MessagingOperation:
    CREATE_QUEUE = "create_queue"
    UPDATE_QUEUE = "update_queue"
    DROP_QUEUE = "drop_queue"
    LIST_QUEUES = "list_queues"
    HAS_QUEUE = "has_queue"
    GET_QUEUE = "get_queue"

    CREATE_TOPIC = "create_topic"
    UPDATE_TOPIC = "update_topic"
    DROP_TOPIC = "drop_topic"
    LIST_TOPICS = "list_topics"
    HAS_TOPIC = "has_topic"
    GET_TOPIC = "get_topic"

    CREATE_SUBSCRIPTION = "create_subscription"
    UPDATE_SUBSCRIPTION = "update_subscription"
    DROP_SUBSCRIPTION = "drop_subscription"
    LIST_SUBSCRIPTIONS = "list_subscriptions"
    HAS_SUBSCRIPTION = "has_subscription"
    GET_SUBSCRIPTION = "get_subscription"

    PUT = "put"
    BATCH = "batch"
    PULL = "pull"
    ACK = "ack"
    NACK = "nack"
    EXTEND = "extend"
    PURGE = "purge"
    CLOSE = "close"



from x8.core import DataModel


class ImageData(DataModel):
    source: str | None = None
    content: bytes | str | None = None
    media_type: str | None = None


class ImageInfo(DataModel):
    width: int
    height: int



from __future__ import annotations

__all__ = ["NextJS"]

import asyncio
import os
import subprocess
import sys
import threading
from pathlib import Path
from typing import Any

from x8.core import Component, Context, Operation, Provider
from x8.interface.api import API


class NextJS(Provider):
    path: str | None
    component: Component | None
    api_host: str | None = None
    api_port: int | None = None
    api_root_path: str | None = None
    api_component: Component | None
    api_url: str | None = None

    _api: API | None

    _template_path = "../templates/nextjs-latest"
    _processes: list[subprocess.Popen]

    def __init__(
        self,
        path: str | None = None,
        component: Component | None = None,
        api_component: Component | None = None,
        api_url: str | None = None,
        api_host: str | None = None,
        api_port: int | None = None,
        api_root_path: str | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            path:
                Path to the web web.
            component:
                Component to run as a web app.
            api_component:
                API component if API component is separately defined.
            api_url:
                API URL if API is already deployed.
            api_host:
                API host.
            api_port:
                API port.
            api_root_path:
                API root path.
        """
        self.path = path
        self.component = component
        self.api_component = api_component
        self.api_url = api_url
        self.api_host = api_host
        self.api_port = api_port
        self.api_root_path = api_root_path

        self._api = None
        self._processes = []

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        server_thread = threading.Thread(target=self._run_api, daemon=True)
        server_thread.start()
        self._put_env_file()
        try:
            self._run_frontend()
        except KeyboardInterrupt:
            sys.exit(1)

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self._put_env_file()
        try:
            await asyncio.gather(
                self._arun_api(),
                self._arun_frontend(),
            )
        except asyncio.CancelledError:
            pass
        except KeyboardInterrupt:
            pass
        finally:
            sys.exit(1)

    def _run_frontend(self):
        shell = os.name == "nt"
        os.chdir(self._get_frontend_path())
        subprocess.run(["npm", "install"], shell=shell, check=True)
        subprocess.run(["npm", "run", "dev"], shell=shell, check=True)

    async def _arun_frontend(self):
        shell = os.name == "nt"
        os.chdir(self._get_frontend_path())

        if shell:
            await asyncio.create_subprocess_shell("npm install")
            await asyncio.create_subprocess_shell("npm run dev")
        else:
            await asyncio.create_subprocess_exec("npm", "install")
            await asyncio.create_subprocess_exec("npm", "run", "dev")

    def _run_api(self):
        api = self._get_api()
        if api:
            api.__run__()

    async def _arun_api(self):
        api = self._get_api()
        if api:
            await api.__arun__()

    def _arun_api_in_thread(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(self._arun_api())
        finally:
            loop.close()

    def _get_api(self) -> API | None:
        component = self.component or self.__component__.component
        if self._api:
            return self._api
        api_component: Any = None
        if self.api_component is not None:
            api_component = self.api_component
        elif component is not None:
            from x8.interface.api import API
            from x8.interface.api.providers.default import Default

            api_component = API(component=component)
            api_component.bind(
                Default(
                    host=self.api_host,
                    port=self.api_port,
                    root_path=self.api_root_path,
                )
            )
        self._api = api_component
        return self._api

    def _get_frontend_path(self) -> str:
        if self.path:
            return self.path
        else:
            return os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__),
                    self._template_path,
                )
            )

    def _put_env_file(self) -> None:
        env_file_path = Path(self._get_frontend_path(), ".env")
        api = self._get_api()
        api_url = "http://127.0.0.1:8080"
        if api:
            api_info = api.get_info().result
            if api_info.port:
                api_url = f"{api_info.host}:{api_info.port}"
            else:
                api_url = f"{api_info.host}"
            if api_info.ssl:
                api_url = f"https://{api_url}"
            else:
                api_url = f"http://{api_url}"
        with env_file_path.open("w") as env_file:
            env_file.write(f"API_URL={api_url}\n")
